# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Test file: test_madlib_observations.py
Description: This file contains unit tests which test "edge cases" in
    the madlib._observations module.
"""

import pathlib
import sys

import astroforge as af
import numpy as np
from astropy.time import Time

# add parent directory of __file__ to sys.path, if isn't already included
if str(pathlib.Path(__file__).parents[1]) not in sys.path:
    sys.path.append(str(pathlib.Path(__file__).parents[1]))

import pytest

import madlib
from madlib._observation import Observation, combineObsCollections
from madlib._utils import MadlibException

# --- setup dummy sensor
SST_PARAMS = {
    "lat": -21.89567,
    "lon": 114.0898731,
    "alt": 0.067845,
    "dra": 0.3,
    "ddec": 0.3,
    "obs_per_collect": (3, 5),
    "obs_time_spacing": 1.5,
    "collect_gap_mean": 7200,
    "collect_gap_std": 800,
    "obs_limits": None,
    "obs_limits": None,
    "id": "SST",
}

seed = 4445


def test_Observation_madlib_exception_1():
    """Observations must be at the same time for computing a residual"""
    with pytest.raises(MadlibException):
        obs1 = Observation(mjd=54321.0)
        obs2 = Observation(mjd=43210.0)
        return obs1 - obs2


def test_Observation_madlib_exception_2():
    """Can only subtract two Observation objects"""
    with pytest.raises(MadlibException):
        obs1 = Observation(mjd=54321.0)
        obs2 = 54321.0
        return obs1 - obs2  # type: ignore


def test_Observation_special_handling_angles_that_wrap():
    obs1 = Observation(mjd=54321.0)
    obs1.ra = 359.0
    obs2 = Observation(mjd=54321.0)
    obs2.ra = 2.0
    residual = obs2 - obs1
    assert np.isclose(residual.ra, 3.0)  # type: ignore


def test_Observation_asarray():
    obs = Observation(mjd=54321.0)
    assert isinstance(obs, madlib._observation.Observation)
    assert isinstance(obs.asarray(), np.ndarray)


def test_ObservationResidual_asarray():
    obs1 = Observation(mjd=54321.0)
    obs2 = Observation(mjd=54321.0)
    residual = obs2 - obs1
    assert isinstance(residual.asarray(), np.ndarray)


def test_ObservationCollection_edge_cases(seed=seed):
    # --- generate an ObservationCollection
    # --- orbital state
    np.random.seed(seed)
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- define satellites
    sat0 = madlib.Satellite(epoch, x0, v0, a0)

    # --- setup sensor and generate ObservationCollection(s)
    sensor = madlib.GroundOpticalSensor(**SST_PARAMS)
    times = sensor.generate_obs_timing(epoch, epoch + 0.5)  # + 0.5 [days]

    ObsColl1 = sensor.observe(sat0, times)
    ObsColl2 = sensor.observe(sat0, times)

    ObsColl1_pos_observed_original_size = ObsColl1.pos_observed.size
    ObsColl1_pos_truth_original_size = ObsColl1.pos_truth.size
    ObsColl1 + ObsColl2  # type: ignore

    collectionList = [ObsColl1, ObsColl2]
    combinedObsCollections = combineObsCollections(collectionList)

    # --- check that ObsColl2 was concatenated with ObsColl1 during the addition line above
    # check if ObsColl1.pos_observed now contains ObsColl2.pos_observed
    assert all(
        ObsColl1.pos_observed[-ObsColl2.pos_observed.size :] == ObsColl2.pos_observed
    )
    assert all(ObsColl1.pos_truth[-ObsColl2.pos_truth.size :] == ObsColl2.pos_truth)
    # check if size is correct
    assert (
        ObsColl1.pos_observed.size
        == ObsColl1_pos_observed_original_size + ObsColl2.pos_observed.size
    )
    assert (
        ObsColl1.pos_truth.size
        == ObsColl1_pos_truth_original_size + ObsColl2.pos_truth.size
    )

    # --- check _obvervation.combineObsCollections()
    assert (
        ObsColl1.pos_observed.size + ObsColl2.pos_observed.size
        == combinedObsCollections.pos_observed.size
    )
    assert (
        ObsColl1.pos_truth.size + ObsColl2.pos_truth.size
        == combinedObsCollections.pos_truth.size
    )



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Test file: test_configs.py
Description: This file contains unit tests for the `configs` module
"""

from madlib._sensor_collection import SensorCollection


def test_sensor_YAML():
    """Test that the sample sensor network YAML file has a valid schema."""
    yaml = "configs/sample_sensor_network.yaml"
    network = SensorCollection.fromYAML(yaml)

    # If the test makes it this far, then the validation check succeeded
    assert True



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import numpy as np
from numpy.typing import NDArray


class MadlibException(Exception):
    """MadlibException Class"""

    pass


sensor_yaml_schema = {
    "type": "object",
    "properties": {
        "sensor_list": {
            "type": "object",
            "patternProperties": {
                ".": {
                    "type": "object",
                    "properties": {
                        "id": {"type": "string"},
                        "lat": {"type": "number", "minimum": -90, "maximum": 90},
                        "lon": {"type": "number", "minimum": -180, "maximum": 360},
                        "alt": {"type": "number"},
                        "dra": {"type": "number", "minimum": 0},
                        "ddec": {"type": "number", "minimum": 0},
                        "students_dof": {"type": "number", "exclusiveMinimum": 0},
                        "obs_per_collect": {"type": "integer", "exclusiveMinimum": 0},
                        "obs_time_spacing": {"type": "number", "minimum": 0},
                        "collect_gap_mean": {"type": "number", "minimum": 0},
                        "collect_gap_std": {"type": "number", "minimum": 0},
                        "obs_limits": {
                            "type": ["object", "null"],
                            "properties": {
                                "el": {
                                    "type": "array",
                                    "items": {
                                        "type": "number",
                                        "minimum": -90,
                                        "maximum": 90,
                                    },
                                    "minItems": 2,
                                    "maxItems": 2,
                                },
                                "az": {
                                    "type": "array",
                                    "items": {
                                        "type": "number",
                                        "minimum": -180,
                                        "maximum": 180,
                                    },
                                    "minItems": 2,
                                    "maxItems": 2,
                                },
                                "sun_el": {
                                    "type": "array",
                                    "items": {
                                        "type": "number",
                                        "minimum": -90,
                                        "maximum": 90,
                                    },
                                    "minItems": 2,
                                    "maxItems": 2,
                                },
                                "dec": {
                                    "type": "array",
                                    "items": {
                                        "type": "number",
                                        "minimum": -90,
                                        "maximum": 90,
                                    },
                                    "minItems": 2,
                                    "maxItems": 2,
                                },
                                "range_": {
                                    "type": "array",
                                    "items": {
                                        "type": "number",
                                        "minimum": 0,
                                    },
                                    "minItems": 2,
                                    "maxItems": 2,
                                },
                            },
                        },
                        "weather": {
                            "type": "object",
                            "properties": {
                                "cloud_prob": {
                                    "type": "number",
                                    "minimum": 0,
                                    "maximum": 1,
                                },
                                "cloud_duration_mean": {"type": "number", "minimum": 0},
                                "cloud_duration_std": {"type": "number", "minimum": 0},
                            },
                        },
                    },
                    "required": [
                        "lat",
                        "lon",
                        "alt",
                        "dra",
                        "ddec",
                        "collect_gap_mean",
                    ],
                    "additionalProperties": False,
                }
            },
        }
    },
    "required": ["sensor_list"],
}


def calc_separation_angle(
    v1: NDArray[np.float64],
    v2: NDArray[np.float64],
    in_deg: bool = False,
):
    """Returns the angle between vectors v1 and v2, both with
    shapes (N, 3). Output is in radians by default, in degrees if
    <in_deg> is True."""

    shape_1 = v1.shape
    shape_2 = v2.shape

    if (len(shape_1) != 2) or (len(shape_2) != 2):
        raise MadlibException("Input arrays must have shape (N,3).")

    if (shape_1[1] != 3) or (shape_2[1] != 3):
        raise MadlibException("Input arrays must have shape (N,3).")

    if shape_1[0] != shape_2[0]:
        raise MadlibException(
            "Input arrays must have the same number of rows. "
            f"The first input has {shape_1[0]} rows, and "
            f"the second input has {shape_2[0]}"
        )

    norm_1 = np.linalg.norm(v1, axis=1, keepdims=True)
    norm_2 = np.linalg.norm(v2, axis=1, keepdims=True)

    v1_u = v1 / norm_1
    v2_u = v2 / norm_2

    dot = np.clip(np.sum(v1_u * v2_u, axis=1), -1.0, 1.0)
    angles = np.arccos(dot)

    if in_deg:
        angles *= 180.0 / np.pi

    return angles



import pytest
import numpy as np

from madlib._utils import calc_separation_angle, MadlibException


def assert_vector_angle(V1, V2, angle_rad):
    sep_rad = calc_separation_angle(V1, V2)
    sep_deg = calc_separation_angle(V1, V2, in_deg=True)

    np.testing.assert_almost_equal(sep_rad, angle_rad, decimal=4)
    np.testing.assert_almost_equal(sep_deg, angle_rad * 180.0 / np.pi, decimal=4)


def assert_multiple_vector_angles(V1, V2, angle_rad_list):
    angle_rad_truth = np.array(angle_rad_list)

    sep_rad = calc_separation_angle(V1, V2)
    sep_deg = calc_separation_angle(V1, V2, in_deg=True)

    np.testing.assert_allclose(sep_rad, angle_rad_truth, atol=1e-4)
    np.testing.assert_allclose(sep_deg, angle_rad_truth * 180.0 / np.pi, atol=1e-4)


class TestSeparationAngle:
    """Test behavior of calc_separation_angle function"""

    def test_single_separations(self):
        """Test the separation between a few individual vectors in radians."""
        N1_1 = np.array(
            [
                [1, 0, 0],
            ]
        )
        N1_2 = np.array(
            [
                [0, 1, 0],
            ]
        )
        N1_3 = np.array(
            [
                [-1, 0, 0],
            ]
        )
        N1_4 = np.array(
            [
                [2, 0, 0],
            ]
        )
        N1_5 = np.array(
            [
                [1, 1, 0],
            ]
        )
        N1_6 = np.array(
            [
                [1, -1, 0],
            ]
        )
        N1_7 = np.array(
            [
                [-1, -1, 0],
            ]
        )
        N1_8 = np.array(
            [
                [1, 2, 3],
            ]
        )

        assert_vector_angle(N1_1, N1_1, 0.0)
        assert_vector_angle(N1_1, N1_2, np.pi / 2)
        assert_vector_angle(N1_2, N1_1, np.pi / 2)
        assert_vector_angle(N1_1, N1_3, np.pi)
        assert_vector_angle(N1_1, N1_4, 0.0)
        assert_vector_angle(N1_1, N1_5, np.pi / 4)
        assert_vector_angle(N1_1, N1_6, np.pi / 4)
        assert_vector_angle(N1_1, N1_7, 3 * np.pi / 4)
        assert_vector_angle(N1_1, N1_8, 1.30024656)

    def test_multiple_vectors(self):
        """Test the pairwise separation between arrays of vectors."""
        N2_1 = np.array(
            [
                [0.1, 0.2, 0.3],
                [0.4, 0.5, 0.6],
                [0.7, 0.8, 0.9],
                [1.0, 1.1, 1.2],
            ]
        )
        N2_2 = np.array(
            [
                [-1.2, -1.1, -1.0],
                [-0.9, -0.8, -0.7],
                [-0.6, -0.5, -0.4],
                [-0.3, -0.2, -0.1],
            ]
        )

        separations = [2.67990488, 2.87801221, 2.87801221, 2.67990488]

        assert_multiple_vector_angles(N2_1, N2_2, separations)

    def test_invalid_shapes(self):
        """Test invalid or incompatible vector shapes."""

        # Case: Neither array is two-dimensional
        B1 = np.array([0, 0, 0])
        B2 = np.array([0, 0, 0])

        with pytest.raises(MadlibException):
            calc_separation_angle(B1, B2)

        # Case: Only one of the arrays is two-dimensional
        B1 = np.array(
            [
                [0, 0, 0],
            ]
        )
        B2 = np.array([0, 0, 0])

        with pytest.raises(MadlibException):
            calc_separation_angle(B1, B2)

        with pytest.raises(MadlibException):
            calc_separation_angle(B2, B1)

        # Case: Neither array has three columns
        B1 = np.array(
            [
                [0, 0],
                [0, 0],
            ]
        )
        B2 = np.array(
            [
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ]
        )

        with pytest.raises(MadlibException):
            calc_separation_angle(B1, B2)

        # Case: Only one of the arrays has three columns
        B1 = np.array(
            [
                [0, 0, 0],
                [0, 0, 0],
            ]
        )
        B2 = np.array(
            [
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ]
        )

        with pytest.raises(MadlibException):
            calc_separation_angle(B1, B2)

        B1 = np.array(
            [
                [0, 0],
                [0, 0],
            ]
        )
        B2 = np.array(
            [
                [0, 0, 0],
                [0, 0, 0],
            ]
        )

        with pytest.raises(MadlibException):
            calc_separation_angle(B1, B2)

        # Case: The arrays have different numbers of rows
        B1 = np.array(
            [
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
            ]
        )
        B2 = np.array(
            [
                [0, 0, 0],
                [0, 0, 0],
            ]
        )

        with pytest.raises(MadlibException):
            calc_separation_angle(B1, B2)



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Test file: test_madlib_sensor.py
Description: This file contains unit tests which test "edge cases" in
    the madlib._sensor module.
"""

import pathlib
import sys

import astroforge as af
import numpy as np
from astropy.time import Time
from astropy.utils import iers

# add parent directory of __file__ to sys.path, if isn't already included
if str(pathlib.Path(__file__).parents[1]) not in sys.path:
    sys.path.append(str(pathlib.Path(__file__).parents[1]))

import madlib
from madlib._utils import MadlibException
from madlib._satellite import Satellite
from madlib._sensor import _Sensor, pos_to_lat_lon
from madlib._observation import Observation

import pytest


seed = 4445


def test_Sensor_NotImplementedErrors():
    with pytest.raises(NotImplementedError):
        _Sensor.generate_obs_timing(None, 54321.0, 54322.0)  # type: ignore

    with pytest.raises(NotImplementedError):
        _epoch = 54321.0
        _pos = np.zeros((3,))
        _vel = np.zeros((3,))
        _accel = np.zeros((3,))
        _manuever_info = None
        sat = Satellite(
            _epoch,
            _pos,
            _vel,
            _accel,
            _manuever_info,
        )
        _Sensor.observe(None, sat, 54321.0)  # type: ignore


def test_invalid_pos_to_lat_lon():
    pos = np.array(
        [
            [8000, 8000, 8000],
            [8000, 8000, 8000],
            [8000, 8000, 8000],
        ]
    )

    times = np.array([1, 2, 3, 4])

    with pytest.raises(ValueError):
        pos_to_lat_lon(pos, times)


class TestGroundOpticalSensor:
    """Test behavior of GroundOpticalSensor objects."""

    def test_GroundOpticalSensor_generate_obs_timing(self, seed=seed):
        np.random.seed(seed)
        # --- orbital state
        th = 2 * np.pi * np.random.rand()
        x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
        v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
        a0 = np.zeros((3,))
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        # --- maneuver definition
        man_time = epoch + 5.0 / 24
        man_dv = np.array([0.0, 5.0, 0.0]) / 1000
        man = madlib.ImpulsiveManeuver(man_time, man_dv)

        # --- setup sensor
        SENSOR_PARAMS = {
            "lat": -21.89567,
            "lon": 114.0898731,
            "alt": 0.067845,
            "dra": 0.3,
            "ddec": 0.3,
            "obs_per_collect": 3,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 7200,
            "collect_gap_std": 800,
            "obs_limits": None,
            "id": "SENSOR",
        }

        sensor = madlib.GroundOpticalSensor(**SENSOR_PARAMS)

        assert not isinstance(sensor.obs_per_collect, (tuple, list))

        with pytest.raises(ValueError):
            sensor.generate_obs_timing(
                epoch, epoch - 2
            )  # - 2 [days], raises ValueError

        assert isinstance(sensor.generate_obs_timing(epoch, epoch + 2), np.ndarray)

    def test_obs_timing_edge(self, seed=seed):
        """If an observation collection contains obs after the simulation end time,
        the entire collection must be discarded."""
        np.random.seed(seed)
        SENSOR_PARAMS = {
            "lat": -21.89567,
            "lon": 114.0898731,
            "alt": 0.067845,
            "dra": 0.3,
            "ddec": 0.3,
            "obs_per_collect": 5,
            "obs_time_spacing": 3 * 3600,
            "collect_gap_mean": 10 * 3600,
            "collect_gap_std": 0,
            "obs_limits": {"el": (20.0, 91.0)},
            "id": "SENSOR",
        }
        sensor = madlib.GroundOpticalSensor(**SENSOR_PARAMS)

        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd
        times = sensor.generate_obs_timing(epoch, epoch + 1)

        # Only 2 of the 3 collects should have succeeded
        assert len(times) == 10

    def test_invalid_limits(self, seed=seed):
        """Test a sensor that contains some invalid observation limits."""
        SENSOR_PARAMS = {
            "lat": -21.89567,
            "lon": 114.0898731,
            "alt": 0.067845,
            "dra": 0.3,
            "ddec": 0.3,
            "obs_per_collect": 1,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 3600,
            "collect_gap_std": 800,
            "obs_limits": {"el": (20.0, 91.0), "el_radians": (0, np.pi)},
            "id": "Sensor",
        }

        # --- generate an ObservationCollection
        # --- orbital state
        np.random.seed(seed)
        th = 2 * np.pi * np.random.rand()
        x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
        v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
        a0 = np.zeros((3,))
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        # --- define satellites
        sat0 = madlib.Satellite(epoch, x0, v0, a0)

        # --- setup sensor and generate ObservationCollection(s)
        sensor = madlib.GroundOpticalSensor(**SENSOR_PARAMS)
        times = sensor.generate_obs_timing(epoch, epoch + 1)

        with pytest.raises(MadlibException):
            obs = sensor.observe(sat0, times)

    def test_invalid_observation(self, seed=seed):
        """Test that limit validation fails on obs without values."""
        SENSOR_PARAMS = {
            "lat": -21.89567,
            "lon": 114.0898731,
            "alt": 0.067845,
            "dra": 0.3,
            "ddec": 0.3,
            "obs_per_collect": 1,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 3600,
            "collect_gap_std": 800,
            "obs_limits": {"el": (20.0, 91.0)},
            "id": "Sensor",
        }
        sensor = madlib.GroundOpticalSensor(**SENSOR_PARAMS)

        obs = Observation(mjd=1000)

        failed = False
        try:
            sensor.validate_limits(obs)
        except MadlibException:
            failed = True

        assert failed

    def test_obs_time_inputs(self):
        """The observe function should still work properly if the
        time input is a single float or a (start, end) tuple."""

        SENSOR_PARAMS = {
            "lat": 0.0,
            "lon": 0.0,
            "alt": 0.0,
            "dra": 0.3,
            "ddec": 0.3,
            "obs_per_collect": 1,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 3600,
            "collect_gap_std": 800,
            "obs_limits": {"el": (20.0, 91.0)},
            "id": "Sensor",
        }
        sensor = madlib.GroundOpticalSensor(**SENSOR_PARAMS)

        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd
        sat = Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)

        obs_1 = sensor.observe(sat, times=epoch + 1)
        obs_2 = sensor.observe(sat, times=(epoch, epoch + 3))

        assert len(obs_1.pos_observed) == 1
        assert len(obs_2.pos_observed) > 1

    def test_truth_position(self):
        """Test the functionality of truth lat/lon/alt positions."""

        SENSOR_1_PARAMS = {
            "lat": -10,
            "lon": 10,
            "alt": 100.0,
            "dra": 0.01,
            "ddec": 0.01,
            "obs_per_collect": 1,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 3600,
            "collect_gap_std": 800,
            "obs_limits": {"el": (20.0, 91.0)},
            "id": "Sensor",
            "lat_truth": 0.0,
            "lon_truth": 0,
            "alt_truth": 0,
        }

        SENSOR_2_PARAMS = {
            "lat": 0,
            "lon": 0,
            "alt": 0.0,
            "dra": 0.01,
            "ddec": 0.01,
            "obs_per_collect": 1,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 3600,
            "collect_gap_std": 800,
            "obs_limits": {"el": (20.0, 91.0)},
            "id": "Sensor",
        }

        sensor_1 = madlib.GroundOpticalSensor(**SENSOR_1_PARAMS)
        sensor_2 = madlib.GroundOpticalSensor(**SENSOR_2_PARAMS)

        assert ~np.isclose(sensor_1.lon, sensor_1.lon_truth, rtol=0, atol=0.1)
        assert ~np.isclose(sensor_1.lat, sensor_1.lat_truth, rtol=0, atol=0.1)
        assert ~np.isclose(sensor_1.alt, sensor_1.alt_truth, rtol=0, atol=0.1)

        assert np.isclose(sensor_2.lon, sensor_2.lon_truth)
        assert np.isclose(sensor_2.lat, sensor_2.lat_truth)
        assert np.isclose(sensor_2.alt, sensor_2.alt_truth)

        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        sat = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        times_1 = sensor_1.generate_obs_timing(epoch, epoch + 1)
        times_2 = sensor_2.generate_obs_timing(epoch, epoch + 1)

        obs_1 = sensor_1.observe(sat, times_1)
        obs_2 = sensor_2.observe(sat, times_2)

        ra_obs_1 = [p.ra for p in obs_1.pos_observed]
        ra_exp_1 = [p.ra for p in obs_1.pos_expected]

        ra_obs_2 = [p.ra for p in obs_2.pos_observed]
        ra_exp_2 = [p.ra for p in obs_2.pos_expected]

        assert all(~np.isclose(ra_obs_1, ra_exp_1, rtol=0, atol=0.1))
        assert all(np.isclose(ra_obs_2, ra_exp_2, rtol=0, atol=0.1))


class TestSpaceOpticalSensor:
    """Test the behavior of SpaceOpticalSensor objects"""

    def test_SpaceOpticalSensor_generate_obs_timing(self, seed=seed):
        np.random.seed(seed)
        # --- orbital state
        th = 2 * np.pi * np.random.rand()
        x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
        v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
        a0 = np.zeros((3,))
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        # --- maneuver definition
        man_time = epoch + 5.0 / 24
        man_dv = np.array([0.0, 5.0, 0.0]) / 1000
        man = madlib.ImpulsiveManeuver(man_time, man_dv)

        # --- setup sensor
        sensor_satellite = madlib.Satellite.from_keplerian(
            epoch=t_start.mjd,
            inclination_rad=0.0,
            raan_rad=0.0,
            argp_rad=0.0,
            ecc=0.0,
            semi_major_axis_km=7000,
            mean_anomaly_rad=0.0,
        )
        SENSOR_PARAMS = {
            "sensor_satellite": sensor_satellite,
            "dra": 0.3,
            "ddec": 0.3,
            "obs_per_collect": 3,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 7200,
            "collect_gap_std": 800,
            "obs_limits": None,
            "id": "SENSOR",
        }

        sensor = madlib.SpaceOpticalSensor(**SENSOR_PARAMS)

        assert not isinstance(sensor.obs_per_collect, (tuple, list))

        with pytest.raises(ValueError):
            sensor.generate_obs_timing(
                epoch, epoch - 2
            )  # - 2 [days], raises ValueError

        assert isinstance(sensor.generate_obs_timing(epoch, epoch + 2), np.ndarray)

    def test_ObservationCollection_no_obs(self, seed=seed):
        # --- setup sensor
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        sensor_satellite = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        SENSOR_PARAMS = {
            "sensor_satellite": sensor_satellite,
            "dra": 0.3,
            "ddec": 0.3,
            "obs_per_collect": 3,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 7200,
            "collect_gap_std": 800,
            "obs_limits": None,
            "id": "SENSOR",
        }

        sensor = madlib.SpaceOpticalSensor(**SENSOR_PARAMS)

        # --- generate an ObservationCollection
        # --- orbital state
        np.random.seed(seed)

        # --- define satellites
        sat0 = madlib.Satellite.from_GEO_longitude(lon=180.0, epoch=epoch)

        # --- setup sensor and generate ObservationCollection(s)
        times = sensor.generate_obs_timing(epoch, epoch + 2)  # + 2 [days]

        ObsColl1 = sensor.observe(sat0, times)
        assert isinstance(ObsColl1, madlib._observation.ObservationCollection)
        assert ObsColl1.count_valid_observations() == 0

    def test_ObservationCollection_no_times(self, seed=seed):
        # --- setup sensor
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        sensor_satellite = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        SENSOR_PARAMS = {
            "sensor_satellite": sensor_satellite,
            "dra": 0.3,
            "ddec": 0.3,
            "obs_per_collect": 3,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 7200,
            "collect_gap_std": 800,
            "obs_limits": None,
            "id": "SENSOR",
        }

        sensor = madlib.SpaceOpticalSensor(**SENSOR_PARAMS)

        # --- generate an ObservationCollection
        # --- orbital state
        np.random.seed(seed)

        # --- define satellites
        sat0 = madlib.Satellite.from_GEO_longitude(lon=1.0, epoch=epoch)

        # --- setup sensor and generate ObservationCollection(s)
        times = (epoch, epoch + 1e-6)

        ObsColl1 = sensor.observe(sat0, times)
        assert isinstance(ObsColl1, madlib._observation.ObservationCollection)
        assert len(ObsColl1.pos_observed) == 0

    def test_ObservationCollection_mistaken_orbit(self, seed=seed):
        # --- setup sensor
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        sensor_reported = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        sensor_truth = madlib.Satellite.from_GEO_longitude(lon=5.0, epoch=epoch)
        SENSOR_PARAMS = {
            "sensor_satellite": sensor_reported,
            "sensor_satellite_truth": sensor_truth,
            "dra": 0.0,
            "ddec": 0.0,
            "obs_per_collect": 3,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 7200,
            "collect_gap_std": 800,
            "obs_limits": None,
            "id": "SENSOR",
        }

        sensor = madlib.SpaceOpticalSensor(**SENSOR_PARAMS)

        # --- generate an ObservationCollection
        # --- orbital state
        np.random.seed(seed)

        # --- define satellites
        sat0 = madlib.Satellite.from_GEO_longitude(lon=1.0, epoch=epoch)

        # --- setup sensor and generate ObservationCollection(s)
        times = (epoch, epoch + 1)

        ObsColl1 = sensor.observe(sat0, times)
        assert isinstance(ObsColl1, madlib._observation.ObservationCollection)

        ra_diff = np.abs(ObsColl1.pos_observed[0].ra - ObsColl1.pos_expected[0].ra)
        assert ra_diff > 100

    def test_ObservationCollection_good_orbit(self, seed=seed):
        # --- setup sensor
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        sensor_reported = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        sensor_truth = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        SENSOR_PARAMS = {
            "sensor_satellite": sensor_reported,
            "sensor_satellite_truth": sensor_truth,
            "dra": 0.0,
            "ddec": 0.0,
            "obs_per_collect": 3,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 7200,
            "collect_gap_std": 800,
            "obs_limits": None,
            "id": "SENSOR",
        }

        sensor = madlib.SpaceOpticalSensor(**SENSOR_PARAMS)

        # --- generate an ObservationCollection
        # --- orbital state
        np.random.seed(seed)

        # --- define satellites
        sat0 = madlib.Satellite.from_GEO_longitude(lon=1.0, epoch=epoch)

        # --- setup sensor and generate ObservationCollection(s)
        times = (epoch, epoch + 1)

        ObsColl1 = sensor.observe(sat0, times)
        assert isinstance(ObsColl1, madlib._observation.ObservationCollection)

        ra_diff = np.abs(ObsColl1.pos_observed[-1].ra - ObsColl1.pos_truth[-1].ra)
        assert ra_diff < 0.01

    def test_observation_single_time(self):
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        sensor_reported = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        sensor_truth = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        SENSOR_PARAMS = {
            "sensor_satellite": sensor_reported,
            "sensor_satellite_truth": sensor_truth,
            "dra": 0.0,
            "ddec": 0.0,
            "obs_per_collect": 3,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 7200,
            "collect_gap_std": 800,
            "obs_limits": None,
            "id": "SENSOR",
        }

        sensor = madlib.SpaceOpticalSensor(**SENSOR_PARAMS)

        sat = madlib.Satellite.from_GEO_longitude(lon=1.0, epoch=epoch)

        obs = sensor.observe(sat, times=epoch + 1)

        assert len(obs.pos_observed) == 1


class TestCrossTags:
    def test_ground_cross_tags(self):
        """Test cross tags on ground-based sensors."""

        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        sat = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        sat_crs = sat.create_cross_tag(
            cross_mjd=epoch + 0.5,
            delta_pos_km=np.array([0, 0, 0]),
            delta_vel_kms=np.array([0, 0, -10]),
        )

        SENSOR_PARAMS = {
            "lat": 0,
            "lon": 0,
            "alt": 0.0,
            "dra": 0.01,
            "ddec": 0.01,
            "obs_per_collect": 1,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 3600,
            "collect_gap_std": 800,
            "obs_limits": {"el": (20.0, 91.0)},
            "id": "Sensor",
            "cross_tag": sat_crs,
            "cross_tag_limit_arcsec": 1000,
        }

        sensor = madlib.GroundOpticalSensor(**SENSOR_PARAMS)
        times = sensor.generate_obs_timing(epoch, epoch + 1)
        obs = sensor.observe(sat, times)

        ra_obs = np.array([p.ra for p in obs.pos_observed])
        ra_exp = np.array([p.ra for p in obs.pos_expected])
        ra_tru = np.array([p.ra for p in obs.pos_truth])

        assert np.sqrt(np.mean((ra_exp - ra_obs) ** 2)) > 10
        assert np.sqrt(np.mean((ra_exp - ra_tru) ** 2)) < 1e-4

    def test_space_cross_tags(self):
        """Test cross-tags on space-based sensors."""

        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        sat = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)
        sat_crs = sat.create_cross_tag(
            cross_mjd=epoch + 0.5,
            delta_pos_km=np.array([0, 0, 0]),
            delta_vel_kms=np.array([0, 0, -10]),
        )

        sensor = madlib.Satellite.from_keplerian(
            epoch=epoch,
            inclination_rad=0,
            raan_rad=0,
            argp_rad=0,
            ecc=0,
            semi_major_axis_km=7000,
            mean_anomaly_rad=0,
        )
        SENSOR_PARAMS = {
            "sensor_satellite": sensor,
            "dra": 0.0,
            "ddec": 0.0,
            "obs_per_collect": 3,
            "obs_time_spacing": 1.5,
            "collect_gap_mean": 7200,
            "collect_gap_std": 800,
            "obs_limits": None,
            "id": "SENSOR",
            "cross_tag": sat_crs,
            "cross_tag_limit_arcsec": 1000,
        }

        sensor = madlib.SpaceOpticalSensor(**SENSOR_PARAMS)

        times = sensor.generate_obs_timing(epoch, epoch + 1)
        obs = sensor.observe(sat, times)

        ra_obs = np.array([p.ra for p in obs.pos_observed])
        ra_exp = np.array([p.ra for p in obs.pos_expected])
        ra_tru = np.array([p.ra for p in obs.pos_truth])

        assert np.sqrt(np.mean((ra_exp - ra_obs) ** 2)) > 10
        assert np.sqrt(np.mean((ra_exp - ra_tru) ** 2)) < 1e-4


def test_pos_to_lat_lon():
    """Test the TETED position to geo lat/lon converter. All angles are in degrees.
    Data used for this test are taken from the JPL Horizons ephemeris of Mars."""

    # If this test is run offline, the dates used in this test are likely too old
    # for the auto IERS file. Disable the max age to prevent an error.
    iers.conf.auto_max_age = None

    times = [
        Time("2024-01-01T00:00:00", format="isot", scale="utc"),
        Time("2024-04-01T00:00:00", format="isot", scale="utc"),
        Time("2024-07-01T00:00:00", format="isot", scale="utc"),
        Time("2024-10-01T00:00:00", format="isot", scale="utc"),
    ]

    # ICRF coordinates (roughly equal to TETED) of Mars computed via JPL Horizons
    x_sun = np.array(
        [
            [-1.907478131413594e07, -3.308234951571553e08, -1.472059527740356e08],
            [2.871238765262319e08, -1.108970588982779e08, -5.534484559072558e07],
            [1.818940970143346e08, 1.722968856880029e08, 6.998787648240359e07],
            [-4.648912080798431e07, 1.641523211579444e08, 7.302598368902390e07],
        ]
    )

    # Precise Mars latitude (declination) and right ascension computed via Horizons
    # (using ICRF coordinates)
    truth_lat = np.array([-23.95204, -10.19497, 15.60632, 23.17272])
    ra = np.array([266.69554, 338.87794, 43.44463, 105.80949])

    # Sub-Mars longitude can be calculated from sidereal time and right ascension
    sidereal_times = np.array(
        [_t.sidereal_time("apparent", "greenwich").value for _t in times]
    )
    hour_angles = (360.0 / 24.0) * sidereal_times - ra
    truth_lon = -hour_angles
    # Wrap angles
    truth_lon[truth_lon < -180] += 360.0
    truth_lon[truth_lon > 180] += 360

    times_mjd = np.array([_t.mjd for _t in times])
    lat, lon = pos_to_lat_lon(x_sun, times_mjd)

    assert all(np.isclose(lat, truth_lat, rtol=0.0, atol=0.01))
    assert all(np.isclose(lon, truth_lon, rtol=0.0, atol=0.01))



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Maneuver Detection Library (MaDLib)
"""

from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

from ._maneuver import ContinuousManeuver, ImpulsiveManeuver
from ._observation import Observation, ObservationResidual
from ._satellite import ContinuousThrustSatellite, Satellite
from ._sensor import GroundOpticalSensor, SpaceOpticalSensor
from ._sensor_collection import SensorCollection

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "MaDDG"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Test file: test_hz_launcher.py
Description: This file contains unit tests for the `hz_launcher` module
"""

import pathlib
import sys
import numpy as np
import pytest

# add parent directory of __file__ to sys.path, if isn't already included
if str(pathlib.Path(__file__).parents[1]) not in sys.path:
    sys.path.append(str(pathlib.Path(__file__).parents[1]))

from scripts.hz_launcher import simulator_task, parseArgs
from madlib._sensor_collection import SensorCollection

sensor_yaml_11 = "configs/sample_sensor_network.yaml"
sensor_params_11 = SensorCollection.paramsFromYAML(sensor_yaml_11)

sensor_yaml_blind = "tests/inputs/blind_sensor.yaml"
sensor_params_blind = SensorCollection.paramsFromYAML(sensor_yaml_blind)


class TestSimulator:
    """Test behavior of simulator_task function"""

    def test_single_non_maneuver(self):
        """Test the outputs of a simple satellite simulation with no maneuvers."""
        seq_id = 0
        maneuver_type = 0
        sim_duration_days = 2.0

        results = simulator_task(
            seq_id=seq_id,
            sensor_params=sensor_params_11,
            maneuver_type=maneuver_type,
            start_mjd=60196.5,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        assert results is not None

        max_mjd = max(results["MJD"])
        min_mjd = min(results["MJD"])

        assert isinstance(max_mjd, float)
        assert isinstance(min_mjd, float)

        # The following assertions must be true for the given inputs
        assert all(results["Maneuver"] == maneuver_type)
        assert all(results["Sequence"] == seq_id)
        assert max_mjd - min_mjd < sim_duration_days
        assert all(results["Maneuver_Start_MJD"].isna())
        assert all(results["Maneuver_End_MJD"].isna())
        assert all(results["Maneuver_DV_Radial_KmS"].isna())
        assert all(results["Maneuver_DV_InTrack_KmS"].isna())
        assert all(results["Maneuver_DV_CrossTrack_KmS"].isna())

    def test_single_maneuver(self):
        """Test the outputs of a simple satellite simulation with an impulsive maneuver."""
        seq_id = 0
        maneuver_type = 1
        sim_duration_days = 2.0

        results = simulator_task(
            seq_id=seq_id,
            sensor_params=sensor_params_11,
            maneuver_type=maneuver_type,
            start_mjd=60196.5,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        assert results is not None

        max_mjd = max(results["MJD"])
        min_mjd = min(results["MJD"])

        assert isinstance(max_mjd, float)
        assert isinstance(min_mjd, float)

        activated_sensors = list(set(results["SensorID"]))

        # The following assertions must be true for the given inputs
        assert all(results["Maneuver"] == maneuver_type)
        assert all(results["Sequence"] == seq_id)
        assert max_mjd - min_mjd < sim_duration_days
        np.testing.assert_allclose(
            results["Maneuver_Start_MJD"], 60197.930379, atol=1e-6
        )
        assert all(results["Maneuver_End_MJD"].isna())
        np.testing.assert_allclose(results["Maneuver_DV_Radial_KmS"], 0, atol=1e-6)
        assert all(np.abs(results["Maneuver_DV_InTrack_KmS"]) > 1e-6)
        assert all(np.abs(results["Maneuver_DV_CrossTrack_KmS"]) > 1e-6)

        # The following assertions must be true for the given seed and default sensors
        assert len(results) == 34
        assert len(activated_sensors) == 3

    def test_start_now(self):
        """Test the no-maneuver case when starting epoch is current time."""
        seq_id = 0
        maneuver_type = 0
        sim_duration_days = 1.0

        results = simulator_task(
            seq_id=seq_id,
            sensor_params=sensor_params_11,
            maneuver_type=maneuver_type,
            start_mjd=None,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        assert results is not None

        max_mjd = max(results["MJD"])
        min_mjd = min(results["MJD"])

        assert isinstance(max_mjd, float)
        assert isinstance(min_mjd, float)

        activated_sensors = list(set(results["SensorID"]))

        # The following assertions must be true for the given inputs
        assert all(results["Maneuver"] == maneuver_type)
        assert all(results["Sequence"] == seq_id)
        assert max_mjd - min_mjd < sim_duration_days

        # The following assertions must be true for the given seed and default sensors
        assert len(results) > 0
        assert len(activated_sensors) > 0

    def test_fixed_continuous(self):
        """Test the outputs of a continuous thrust with a fixed vector."""
        seq_id = 0
        maneuver_type = 2
        cont_thrust_duration_days = 0.25
        cont_thrust_model = 0
        cont_thrust_mag = 1e-9
        sim_duration_days = 2.0

        results = simulator_task(
            seq_id=seq_id,
            sensor_params=sensor_params_11,
            maneuver_type=maneuver_type,
            cont_thrust_duration_days=cont_thrust_duration_days,
            cont_thrust_mag=cont_thrust_mag,
            cont_thrust_model=cont_thrust_model,
            start_mjd=60196.5,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        assert results is not None

        max_mjd = max(results["MJD"])
        min_mjd = min(results["MJD"])

        assert isinstance(max_mjd, float)
        assert isinstance(min_mjd, float)

        activated_sensors = list(set(results["SensorID"]))

        final_ra = results["RA Arcsec"].iloc[-1]
        final_dec = results["DEC Arcsec"].iloc[-1]

        # The following assertions must be true for the given inputs
        assert all(results["Maneuver"] == maneuver_type)
        assert all(results["Sequence"] == seq_id)
        assert max_mjd - min_mjd < sim_duration_days
        np.testing.assert_allclose(results["Maneuver_Start_MJD"], 60196.5, atol=1e-6)
        np.testing.assert_allclose(results["Maneuver_End_MJD"], 60196.75, atol=1e-6)
        np.testing.assert_allclose(results["Maneuver_DV_Radial_KmS"], 0, atol=1e-6)
        np.testing.assert_allclose(
            results["Maneuver_DV_InTrack_KmS"], cont_thrust_mag, atol=1e-6
        )
        np.testing.assert_allclose(results["Maneuver_DV_CrossTrack_KmS"], 0, atol=1e-6)

    def test_random_continuous(self):
        """Test the outputs of a continuous thrust with a random vector."""
        seq_id = 0
        maneuver_type = 2
        cont_thrust_mag = 1e-9
        cont_thrust_duration_days = 0.5
        cont_thrust_model = 1
        sim_duration_days = 2.0

        results = simulator_task(
            seq_id=seq_id,
            sensor_params=sensor_params_11,
            maneuver_type=maneuver_type,
            cont_thrust_mag=cont_thrust_mag,
            cont_thrust_duration_days=cont_thrust_duration_days,
            cont_thrust_model=cont_thrust_model,
            start_mjd=60196.5,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        assert results is not None

        max_mjd = max(results["MJD"])
        min_mjd = min(results["MJD"])

        assert isinstance(max_mjd, float)
        assert isinstance(min_mjd, float)

        # The following assertions must be true for the given inputs
        assert all(results["Maneuver"] == maneuver_type)
        assert all(results["Sequence"] == seq_id)
        assert max_mjd - min_mjd < sim_duration_days
        np.testing.assert_allclose(results["Maneuver_Start_MJD"], 60196.5, atol=1e-6)
        # By default, continuous mnvr duration should be equal to simulation duration
        np.testing.assert_allclose(results["Maneuver_End_MJD"], 60197.0, atol=1e-6)
        np.testing.assert_allclose(
            results["Maneuver_DV_Radial_KmS"], 2.605e-10, atol=1e-12
        )
        np.testing.assert_allclose(
            results["Maneuver_DV_InTrack_KmS"], 5.454e-10, atol=1e-12
        )
        np.testing.assert_allclose(
            results["Maneuver_DV_CrossTrack_KmS"], -7.967e-10, atol=1e-12
        )

    def test_invalid_continuous(self):
        """Test the outputs of a continuous thrust with an invalid thrust mode."""
        seq_id = 0
        maneuver_type = 2
        cont_thrust_model = 3
        sim_duration_days = 0.5

        failed = False
        try:
            results = simulator_task(
                seq_id=seq_id,
                sensor_params=sensor_params_11,
                maneuver_type=maneuver_type,
                cont_thrust_model=cont_thrust_model,
                start_mjd=60196.5,
                sim_duration_days=sim_duration_days,
                random_seed=0,
            )
        except ValueError:
            failed = True

        assert failed

    def test_invalid_thrust_mode(self):
        """Test that an invalid thrust mode returns None"""
        seq_id = 0
        maneuver_type = 3
        cont_thrust_model = 1
        sim_duration_days = 0.5

        results = simulator_task(
            seq_id=seq_id,
            sensor_params=sensor_params_11,
            maneuver_type=maneuver_type,
            cont_thrust_model=cont_thrust_model,
            start_mjd=60196.5,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        assert results == None

    def test_no_obs(self):
        """Test that a simulation with no observations returns None"""
        seq_id = 0
        maneuver_type = 0
        sim_duration_days = 0.5

        results = simulator_task(
            seq_id=seq_id,
            sensor_params=sensor_params_blind,
            maneuver_type=maneuver_type,
            start_mjd=60196.5,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        assert results is None


class TestArgs:
    """Test behavior of argument parsing"""

    def test_defaults(self):
        """Test default argument values"""
        inputs = ["0", "sensor_yaml_path"]

        parser = parseArgs()
        args = parser.parse_args(inputs)

        # Count number of input arguments
        # (This is to make sure you update the tests when you add an arg)
        assert len(args.__dict__) == 19

        assert isinstance(args.num_pairs, int)
        assert args.num_pairs == 0
        assert isinstance(args.sensor_yaml, str)
        assert args.sensor_yaml == "sensor_yaml_path"
        assert isinstance(args.start_mjd, float)
        np.testing.assert_almost_equal(args.start_mjd, -1, decimal=7)
        assert isinstance(args.sim_duration_days, float)
        np.testing.assert_almost_equal(args.sim_duration_days, 3.0, decimal=7)
        assert isinstance(args.mtype, str)
        assert args.mtype == "impulse"
        assert args.cont_thrust_duration_days is None
        assert isinstance(args.cont_thrust_model, int)
        assert args.cont_thrust_model == 0
        assert isinstance(args.cont_thrust_mag, float)
        np.testing.assert_almost_equal(args.cont_thrust_mag, 1e-7, decimal=9)
        assert isinstance(args.outdir, str)
        assert args.outdir == "outputs"
        assert isinstance(args.submitit, str)
        assert args.submitit == ""
        assert isinstance(args.multirun_root, str)
        assert args.multirun_root == ""
        assert isinstance(args.pred_err, float)
        np.testing.assert_almost_equal(args.pred_err, 0, decimal=7)
        assert isinstance(args.rm_multirun_root, bool)
        assert not args.rm_multirun_root
        assert isinstance(args.overwrite, bool)
        assert not args.overwrite

        dv_mean = args.dv_ric_mean_kms
        dv_std = args.dv_ric_std_kms

        assert isinstance(dv_mean, list)
        assert isinstance(dv_std, list)
        assert len(dv_mean) == 3
        assert len(dv_std) == 3
        np.testing.assert_almost_equal(dv_mean[0], 0, decimal=7)
        np.testing.assert_almost_equal(dv_mean[1], 0, decimal=7)
        np.testing.assert_almost_equal(dv_mean[2], 0, decimal=7)
        np.testing.assert_almost_equal(dv_std[0], 0, decimal=7)
        np.testing.assert_almost_equal(dv_std[1], 0.1, decimal=7)
        np.testing.assert_almost_equal(dv_std[2], 1, decimal=7)

    def test_inputs(self):
        """Test that input values work as expected"""
        inputs = [
            "10",
            "sensor_yaml_path",
            "--start_mjd",
            "1000",
            "--sim_duration_days",
            "5.0",
            "--mtype",
            "continuous",
            "--dv_ric_mean_kms",
            "2",
            "2",
            "2",
            "--dv_ric_std_kms",
            "5",
            "5",
            "5",
            "--sensor_dra",
            "1",
            "--sensor_ddec",
            "1",
            "--cont_thrust_duration_days",
            "2.5",
            "--cont_thrust_model",
            "1",
            "--cont_thrust_mag",
            "1e-3",
            "--outdir",
            "test1",
            "--submitit",
            "test2",
            "--multirun_root",
            "test3",
            "--pred_err",
            "1e-3",
            "--rm_multirun_root",
            "--overwrite",
            "--sims_per_task",
            "10",
        ]

        parser = parseArgs()
        args = parser.parse_args(inputs)

        # Count number of input arguments
        # (This is to make sure you update the tests when you add an arg)
        assert len(args.__dict__) == 19

        assert isinstance(args.num_pairs, int)
        assert isinstance(args.sensor_yaml, str)
        assert args.sensor_yaml == "sensor_yaml_path"
        assert args.num_pairs == 10
        assert isinstance(args.start_mjd, float)
        np.testing.assert_almost_equal(args.start_mjd, 1000, decimal=7)
        assert isinstance(args.sim_duration_days, float)
        np.testing.assert_almost_equal(args.sim_duration_days, 5.0, decimal=7)
        assert isinstance(args.mtype, str)
        assert args.mtype == "continuous"
        assert isinstance(args.sensor_dra, float)
        np.testing.assert_almost_equal(args.sensor_dra, 1.0, decimal=7)
        assert isinstance(args.sensor_ddec, float)
        np.testing.assert_almost_equal(args.sensor_ddec, 1.0, decimal=7)
        assert isinstance(args.cont_thrust_duration_days, float)
        np.testing.assert_almost_equal(args.cont_thrust_duration_days, 2.5)
        assert isinstance(args.cont_thrust_model, int)
        assert args.cont_thrust_model == 1
        assert isinstance(args.cont_thrust_mag, float)
        np.testing.assert_almost_equal(args.cont_thrust_mag, 1e-3, decimal=9)
        assert isinstance(args.outdir, str)
        assert args.outdir == "test1"
        assert isinstance(args.submitit, str)
        assert args.submitit == "test2"
        assert isinstance(args.multirun_root, str)
        assert args.multirun_root == "test3"
        assert isinstance(args.pred_err, float)
        np.testing.assert_almost_equal(args.pred_err, 1e-3, decimal=9)
        assert isinstance(args.rm_multirun_root, bool)
        assert args.rm_multirun_root
        assert isinstance(args.overwrite, bool)
        assert args.overwrite
        assert isinstance(args.sims_per_task, int)
        assert args.sims_per_task == 10

        dv_mean = args.dv_ric_mean_kms
        dv_std = args.dv_ric_std_kms

        assert isinstance(dv_mean, list)
        assert isinstance(dv_std, list)
        assert len(dv_mean) == 3
        assert len(dv_std) == 3
        np.testing.assert_almost_equal(dv_mean[0], 2, decimal=7)
        np.testing.assert_almost_equal(dv_mean[1], 2, decimal=7)
        np.testing.assert_almost_equal(dv_mean[2], 2, decimal=7)
        np.testing.assert_almost_equal(dv_std[0], 5, decimal=7)
        np.testing.assert_almost_equal(dv_std[1], 5, decimal=7)
        np.testing.assert_almost_equal(dv_std[2], 5, decimal=7)

    class TestRicLengths:
        """Make sure the RIC inputs can only have length 3"""

        def test_mean_short(self):
            """dv_ric_mean must have 3 inputs"""
            inputs = [
                "10",
                "sensor_yaml_path",
                "--dv_ric_mean_kms",
                "2",
                "2",
            ]

            parser = parseArgs()

            with pytest.raises(SystemExit):
                parser.parse_args(inputs)

        def test_mean_long(self):
            """dv_ric_mean must have 3 inputs"""
            inputs = [
                "10",
                "sensor_yaml_path",
                "--dv_ric_mean_kms",
                "2",
                "2",
                "2",
                "2",
            ]

            parser = parseArgs()

            with pytest.raises(SystemExit):
                parser.parse_args(inputs)

        def test_std_short(self):
            """dv_ric_std must have 3 inputs"""
            inputs = [
                "10",
                "sensor_yaml_path",
                "--dv_std_mean_kms",
                "2",
                "2",
            ]

            parser = parseArgs()

            with pytest.raises(SystemExit):
                parser.parse_args(inputs)

        def test_std_long(self):
            """dv_ric_std must have 3 inputs"""
            inputs = [
                "10",
                "sensor_yaml_path",
                "--dv_std_mean_kms",
                "2",
                "2",
                "2",
                "2",
            ]

            parser = parseArgs()

            with pytest.raises(SystemExit):
                parser.parse_args(inputs)



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import argparse
import logging
import pathlib
import shutil
import sys
import time
import warnings
from typing import Tuple

import numpy as np
import pandas as pd
from astropy.time import Time
from hydra.conf import HydraConf, JobConf
from hydra_zen import store

import madlib
from maddg._residuals import calculate_residuals
from maddg._sim_launcher import launcher
from madlib._utils import MadlibException

log_task = logging.getLogger("simulator_task")
log_task.setLevel(logging.WARNING)


def parseArgs():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        "num_pairs",
        type=int,
        help="Number of Maneuver/Non-Maneuver pairs per simulation launch",
    )

    parser.add_argument(
        "sensor_yaml",
        type=str,
        help="Path to a YAML file defining the sensor network for the simulation",
    )

    parser.add_argument(
        "--start_mjd",
        type=float,
        default=-1.0,
        help="MJD at start of simulation. If <0, start at current system time. (default: -1.0)",
    )

    parser.add_argument(
        "--sim_duration_days",
        type=float,
        default=3.0,
        help=("Duration of simulation in whole days\n" "  (default: %(default)f)"),
    )

    parser.add_argument(
        "--mtype",
        type=str,
        choices=["impulse", "continuous", "all"],
        default="impulse",
        help=("Maneuver type for orbit simulations\n" "  (default: %(default)s)"),
    )

    parser.add_argument(
        "--dv_ric_mean_kms",
        type=float,
        nargs=3,
        default=[0, 0, 0],
        help=(
            "Mean values for impulsive RIC thrust vector normal distributions, in km/s\n"
            "  (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "--dv_ric_std_kms",
        type=float,
        nargs=3,
        default=[0, 0.1, 1],
        help=(
            "Standard deviations for impulsive RIC thrust vector normal distributions, in km/s\n"
            "  (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "--sensor_dra",
        type=float,
        default=None,
        help=(
            "Sensor metric accuracy in the right ascension direction (arcsec).\n"
            "  If not set, value is None, and `dra` value in sensor_yaml file\n"
            "  will be used. (default: None)"
        ),
    )

    parser.add_argument(
        "--sensor_ddec",
        type=float,
        default=None,
        help=(
            "Sensor metric accuracy in the declination direction (arcsec).\n"
            "  If not set, value is None, and `dra` value in sensor_yaml file\n"
            "  will be used. (default: None)"
        ),
    )

    parser.add_argument(
        "--cont_thrust_duration_days",
        type=float,
        help=(
            "Duration in days of continuous thrust, which begins at start of simulation\n"
            "  (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "--cont_thrust_model",
        type=int,
        choices=[0, 1],
        default=0,
        help=(
            "Continous thrust model:\n"
            "  0 = continuous thrust in [0,1,0] direction\n"
            "  1 = continuous thrust in random direction\n"
            "  (relevant only for --mtype continuous)\n"
            "  (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "--cont_thrust_mag",
        type=float,
        default=1.0e-7,
        help=(
            "Magnitude of the constant acceleration (km/s^2) for any continuous constant trust models\n"
            "  (relevant only for --mtype continuous, --cont_thrust_mode {0, 1})\n"
            "  (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "--outdir",
        type=str,
        default="outputs",
        help="Path to output directory",
    )

    parser.add_argument(
        "--submitit",
        type=str,
        default="",
        help="Optional JSON config file defining how to launch jobs across multiple GPUs using submitit",
    )

    parser.add_argument(
        "--multirun_root",
        type=str,
        default="",
        help="Optional path to directory where multirun results should be saved (./multirun by default)",
    )

    parser.add_argument(
        "--pred_err",
        type=float,
        default=0.0,
        help="Fractional error on predicted initial orbital state",
    )

    parser.add_argument(
        "--rm_multirun_root",
        action="store_true",
        help="Raise this flag to remove the multirun root directory after merging data",
    )

    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Raise this flag to proceed with overwriting any data in outdir if outdir exists and is not empty",
    )

    parser.add_argument(
        "--sims_per_task",
        type=int,
        default=1,
        help=(
            "Number of simulations per task function. Must evenly divide into `num_pairs`.\n"
            "  (default: %(default)i)"
        ),
    )

    return parser


def simulator_task(
    seq_id: int,
    sensor_params: dict,
    maneuver_type: int,
    sim_duration_days: float,
    num_sim_pairs: int | None = None,
    start_mjd: float | None = None,
    dv_ric_mean_kms: Tuple[float, float, float] = (0.0, 0.0, 0.0),
    dv_ric_std_kms: Tuple[float, float, float] = (0.0, 0.1, 1.0),
    cont_thrust_duration_days: float | None = None,
    cont_thrust_mag: float = 1e-7,
    cont_thrust_model: int = 0,
    pred_err: float = 0.0,
    random_seed: int | None = None,
    sims_per_task: int = 1,
) -> pd.DataFrame | None:
    """Generates a satellite, propagates it with and without a maneuver.

    Parameters
    ----------
    seq_id : int
        Unique ID of satellite
    sensor_params: dict
        Dictionary defining the parameters for each active sensor
    maneuver_type : int
        Specifies whether ot not the satellite is performing a maneuver:
        0 = no maneuver,
        1 = impulse maneuver,
        2 = continuous maneuver,
    sim_duration_days : float
        Duration of the simulation (days)
    num_sim_pairs : int, optional
        Number of simulation to perform per maneuver type.
    start_mjd : float, optional
        MJD at which the simulation should begin, by default None (current MJD)
    dv_ric_mean_kms : Tuple[float, float, float], optional
        Mean values of normal distributions to use when sampling
        the radial, in-track, and cross-track delta-V values, respectively,
        of impulsive maneuvers. In units of km/s, by default (0.0, 0.0, 0.0)
    dv_ric_std_kms : Tuple[float, float, float], optional
        Standard deviations of normal distributions to use when sampling
        the radial, in-track, and cross-track delta-V values, respectively,
        of impulsive maneuvers. In units of km/s, by default (0.0, 0.1, 1.0)
    cont_thrust_duration_days : float | None, optional
        Duration in days of the continuous maneuver that begins at simulation start, by default None
        (if None, maneuver duration is equal to simulation duration)
        Magnitude of the continuous thrust (km/s/s) (default 1e-7)
    cont_thrust_model : int, optional
        Which continuous thrust model to use:
        0 = applies a continuous thrust in the [0,1,0] direction,
        1 = applies a continuous thrust in a random direction,
        (default 0)
    pred_err : float, optional
        Fractional error on predicted initial orbital state, by default 0.0
    random_seed : int, optional
        Random seed to use for numpy, by default None
    sims_per_task : int, optional
        Number of simulations to perform per task function, by default 1

    Returns
    -------
    pd.core.frame.DataFrame
        Results (residuals)
    or, None (if no result)

    Raises
    ------
    ValueError
        Occurs if supplied `cont_thrust_model` is not a supported option
    """
    # Set the random seed for numpy
    np.random.seed(seed=random_seed)

    # Offset `seq_id` based off of value of `maneuver_type` so that each sequence ID is unique
    if num_sim_pairs is not None:
        seq_id = seq_id + num_sim_pairs * maneuver_type

    residual_dfs = []
    for seq_id in np.arange(seq_id, seq_id + sims_per_task):
        if cont_thrust_duration_days is None:
            cont_thrust_duration_days = sim_duration_days

        # Declarations
        sat_observed = None

        sensors = [
            madlib.GroundOpticalSensor(**params)
            for key, params in sensor_params.items()
        ]
        sensor_network = madlib.SensorCollection(sensors)

        # Timing
        if start_mjd is None:
            epoch = Time.now().mjd + np.random.random()
        else:
            epoch = start_mjd

        # Satellite Parameters
        sat_longitude = 360 * np.random.random()

        # Define Maneuver and setup appropriate Satellite Class
        maneuver = None
        maneuver_start_mjd = None
        maneuver_end_mjd = None
        maneuver_r_kms = None
        maneuver_i_kms = None
        maneuver_c_kms = None
        if maneuver_type == 0:
            """no maneuver"""
            # Setup Satellite
            sat_observed = madlib.Satellite.from_GEO_longitude(sat_longitude, epoch)
        elif maneuver_type == 1:
            """impulse maneuver"""
            # Maneuver Definition
            man_time = epoch + sim_duration_days * np.random.random()
            maneuver_start_mjd = man_time

            mean_rad, mean_in, mean_crs = dv_ric_mean_kms
            std_rad, std_in, std_crs = dv_ric_std_kms
            dv_rad = mean_rad + std_rad * np.random.randn()
            dv_in = mean_in + std_in * np.random.randn()
            dv_crs = mean_crs + std_crs * np.random.randn()

            man_dv = np.array([dv_rad, dv_in, dv_crs]) / 1000
            maneuver = madlib.ImpulsiveManeuver(man_time, man_dv)

            maneuver_r_kms = dv_rad
            maneuver_i_kms = dv_in
            maneuver_c_kms = dv_crs
            # Setup Satellite
            sat_observed = madlib.Satellite.from_GEO_longitude(sat_longitude, epoch)
        elif maneuver_type == 2:
            """continuous maneuver"""

            if cont_thrust_model == 0:
                accel_vec = np.array([0.0, 1.0, 0.0]) * cont_thrust_mag

                def acc_f(t):
                    return accel_vec

            elif cont_thrust_model == 1:
                rand_unit_vec = np.random.randn(3)
                rand_unit_vec /= np.linalg.norm(rand_unit_vec)
                accel_vec = rand_unit_vec * cont_thrust_mag

                def acc_f(t):
                    return accel_vec

            else:
                raise ValueError(f"{cont_thrust_model = } is not a supported option")

            man_time = (epoch, epoch + cont_thrust_duration_days)
            maneuver = madlib.ContinuousManeuver(acc_f, man_time)
            # Setup Satellite
            sat_observed = madlib.ContinuousThrustSatellite.from_GEO_longitude(
                sat_longitude, epoch
            )

            maneuver_start_mjd = man_time[0]
            maneuver_end_mjd = man_time[1]
            maneuver_r_kms, maneuver_i_kms, maneuver_c_kms = accel_vec

        if isinstance(
            sat_observed, (madlib.Satellite, madlib.ContinuousThrustSatellite)
        ):
            # If prediction error has been specified, use it to define a "true" orbit
            if np.abs(pred_err) > 1e-8:
                r_err = 1 + pred_err * np.random.randn(3)
                v_err = 1 + pred_err * np.random.randn(3)

                true_r = sat_observed.x * r_err
                true_v = sat_observed.v * v_err

                sat_observed.x_true = true_r
                sat_observed.v_true = true_v

            sat_observed.maneuver = maneuver

            # Observe and calculate residuals
            residual_df = calculate_residuals(
                sensors=sensor_network,
                satellite=sat_observed,
                sim_duration_days=sim_duration_days,
                t_start_mjd=epoch,
            )

            if residual_df is not None:
                residual_df["Maneuver"] = maneuver_type
                residual_df["Sequence"] = int(seq_id)
                residual_df["Maneuver_Start_MJD"] = maneuver_start_mjd
                residual_df["Maneuver_End_MJD"] = maneuver_end_mjd
                residual_df["Maneuver_DV_Radial_KmS"] = maneuver_r_kms
                residual_df["Maneuver_DV_InTrack_KmS"] = maneuver_i_kms
                residual_df["Maneuver_DV_CrossTrack_KmS"] = maneuver_c_kms
                residual_dfs.append(residual_df)

    if len(residual_dfs) == 0:
        return None
    else:
        return pd.concat(residual_dfs, ignore_index=True)


if __name__ == "__main__":
    time_start = time.time()

    # parse arguments
    parser = parseArgs()
    args = parser.parse_args()

    # check if num_pairs is evenly divisible by sims_per_task
    if args.num_pairs % args.sims_per_task != 0:
        warnings.warn(
            "`num_pairs` is not evenly divisible by `sims_per_task`. Number of sequences generated may be different than expected."
        )

    # check outdir for any existing files
    try:
        outdir_has_files = any(pathlib.Path(args.outdir).iterdir())
    except FileNotFoundError:
        outdir_has_files = False

    # exit (abort simulation) if outdir is not empty, unless --overwrite was specified
    # otherwise, delete existing folder and proceed
    if outdir_has_files and (not args.overwrite):
        # abort simulation (do nothing)
        sys.exit(
            "Outdir exists and is not empty. --overwrite argument was not passed. Aborting simulation."
        )
    elif outdir_has_files and args.overwrite:
        # delete all files in outdir and proceed
        shutil.rmtree(args.outdir)
        print(
            "INFO :: outdir exists and --overwrite argument was passed. Existing data was deleted."
        )

    # customize Hydra's configuration
    store(HydraConf(job=JobConf(chdir=True)))
    store.add_to_hydra_store(overwrite_ok=True)

    simulator_method = simulator_task
    num_sim_pairs = args.num_pairs
    sensor_yaml = args.sensor_yaml
    outdir = args.outdir
    dv_ric_mean_kms = (
        args.dv_ric_mean_kms[0],
        args.dv_ric_mean_kms[1],
        args.dv_ric_mean_kms[2],
    )
    dv_ric_std_kms = (
        args.dv_ric_std_kms[0],
        args.dv_ric_std_kms[1],
        args.dv_ric_std_kms[2],
    )

    sim_duration_days = args.sim_duration_days

    if not args.cont_thrust_duration_days:
        cont_thrust_duration_days = sim_duration_days
    else:
        cont_thrust_duration_days = args.cont_thrust_duration_days

    cont_thrust_mag = args.cont_thrust_mag
    cont_thrust_model = args.cont_thrust_model
    pred_err = args.pred_err
    submitit = args.submitit
    multirun_root = args.multirun_root
    rm_multirun_root = args.rm_multirun_root
    mtype = args.mtype
    sensor_dra = args.sensor_dra
    sensor_ddec = args.sensor_ddec

    start_mjd = args.start_mjd
    if start_mjd < 0:
        start_mjd = None

    print("Setting up job launcher...")

    launcher(
        simulator_method,
        mtype,
        num_sim_pairs,
        sensor_yaml,
        outdir,
        dv_ric_mean_kms=dv_ric_mean_kms,
        dv_ric_std_kms=dv_ric_std_kms,
        cont_thrust_duration_days=cont_thrust_duration_days,
        cont_thrust_mag=cont_thrust_mag,
        cont_thrust_model=cont_thrust_model,
        submitit=submitit,
        multirun_root=multirun_root,
        rm_multirun_root=rm_multirun_root,
        start_mjd=start_mjd,
        sim_duration_days=sim_duration_days,
        pred_err=pred_err,
        sensor_dra=sensor_dra,
        sensor_ddec=sensor_ddec,
        sims_per_task=args.sims_per_task,
    )
    time_stop = time.time()
    print(f"INFO :: Elapsed time (sec) = {time_stop - time_start}")



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import argparse
import json
from pathlib import Path
from typing import Any, Callable, Dict, List

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy.time import Time
from hydra.conf import HydraConf, JobConf
from hydra_zen import launch, make_config, store, zen
from hydra_zen.typing._implementations import DataClass

from madlib._utils import MadlibException

from madlib import SensorCollection

plt.rcParams.update({"figure.max_open_warning": 0})


def parseArgs():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Script description",
    )
    # Add the required arguments
    parser.add_argument(
        "--path",
        type=str,
        required=True,
        help="The path to the input data file (.csv)",
    )
    parser.add_argument(
        "--sensor_yaml",
        type=str,
        required=True,
        help="Path to a YAML file defining the sensor network for the simulation",
    )
    # Optional arguments
    parser.add_argument(
        "--cloud_prob",
        type=str,
        default="0.5",
        help="Probability of a cloud event blocking the sky during any nighttime observable window for each sensor",
    )
    parser.add_argument(
        "--cloud_duration_mean",
        type=str,
        default="10800.0",
        help="Mean duration of a cloud event (seconds)",
    )
    parser.add_argument(
        "--cloud_duration_std",
        type=str,
        default="3600.0",
        help="Standard deviation of a cloud event (seconds)",
    )
    parser.add_argument(
        "--num_runs",
        type=int,
        default=1,
        help="Number of dropout datasets to generate",
    )
    parser.add_argument(
        "--save_copy_of_original",
        action="store_true",
        help="Raise this flag to save copy of the original input data along side of the modified dataset with dropouts",
    )
    parser.add_argument(
        "--save_plots",
        action="store_true",
        help="Raise this flag to generate and save plots",
    )
    parser.add_argument(
        "--submitit",
        type=str,
        default="",
        help="Optional JSON config file defining how to launch jobs across multiple GPUs using submitit",
    )

    return parser


def gen_random_duration(
    loc: float,
    scale: float,
) -> float:
    """Generate a random weather duration, drawing from the random normal distribution.

    Parameters
    ----------
    loc : float
        Mean ("center") of the distrubtion.
    scale : float
        Standard deviation (spread or "width") of the distribution. Must be non-negative.

    Returns
    -------
    float
        Random weather duration (note: always positive)
    """
    duration = -1.0
    # generate a duration sample until it is non-negative:
    while duration < 0.0:
        duration = np.random.normal(loc=loc, scale=scale)
    return duration


def get_sensor(
    sensor_list: List[Dict[str, Any]],
    sensor_ID: str,
) -> Dict[str, Any] | None:
    """Get sensor dictionary that corresponds to sensor_ID.

    Parameters
    ----------
    sensor_list : List[Dict[str, Any]]
        List of sensors
    sensor_ID : str
        Dictionary within sensor_list with a sensor_ID that matches this parameter will be returned.

    Returns
    -------
    Dict[str, Any] | None
    """
    sensor_match = None
    for sensor in sensor_list:
        if sensor["id"] == sensor_ID:
            sensor_match = sensor
            break
    return sensor_match


def update_sensor_weather_stats(
    sensor_list: List[Dict[str, Any]],
    idx: int | None = None,
    cloud_prob: float = 0.5,
    cloud_duration_mean: float = 10800.0,
    cloud_duration_std: float = 7200.0,
) -> List[Dict[str, Any]]:
    """Updates sensor weather stats for the sensors in supplied sensor_list.

    Parameters
    ----------
    sensor_list : List[Dict[str, Any]]
        List of sensors
    idx : int | None, optional
        Sensor index, by default None. If None, all sensors in the sensor list will be updated.
    cloud_prob : float, optional
       Probability of a cloud blocking a measurement: 0 - 1, by default 0.5
    cloud_duration_mean : float, optional
        Average duration of each cloud event (seconds), by default 10800.0
    cloud_duration_std : float, optional
        Standard deviation of each cloud event (seconds), by default 7200.0

    Returns
    -------
    List[Dict[str, Any]]
        _description_
    """
    s_l = sensor_list.copy()
    if isinstance(idx, int):
        s_l[idx]["weather"]["cloud_prob"] = cloud_prob
        s_l[idx]["weather"]["cloud_duration_mean"] = cloud_duration_mean
        s_l[idx]["weather"]["cloud_duration_std"] = cloud_duration_std
    else:
        for sensor in s_l:
            sensor["weather"]["cloud_prob"] = cloud_prob
            sensor["weather"]["cloud_duration_mean"] = cloud_duration_mean
            sensor["weather"]["cloud_duration_std"] = cloud_duration_std
    return s_l


def plot_sensor_ID(
    sensor_ID: str,
    df: pd.DataFrame,
    df_drop: pd.DataFrame,
    sensor_list: List[Dict[str, Any]],
    save_without_displaying: bool = False,
    save_path: str = "./plots/sensor_ID/",
) -> None:
    """This function plots all measurements for a given sensor_ID, before and after a synthetic "weather event".
    The data dropout regions are highlighted in the after plot.

    Parameters
    ----------
    sensor_ID : str
        The unique sensor ID
    df : pd.DataFrame
        The pandas DataFrame of a MaDDG output file, with an additional ``Datetime`` column
    df_drop : pd.DataFrame
        The pandas DataFrame of a MaDDG output file, with mising data, and an additional ``Datetime`` column
    sensor_list : List[Dict[str, Any]]
        List of sensors
    save_without_displaying : bool, optional
        Whether or not to save plots to the hydra multirun directory, without displaying the plots during runtime, by default False
    save_path : str, optional
        Path to save the plots within the hydra multirun directory, by default "./plots/sensor_ID/"
    """
    sensor = get_sensor(sensor_list, sensor_ID)
    if sensor is not None:
        try:
            hours = mdates.HourLocator(interval=8)
            timeFmt = mdates.DateFormatter("%H:")
            m_size = 4
            _ax_idx = [0, 2]
            _plt_types = ["original", "with dropout"]
            fig, axs = plt.subplots(nrows=4, ncols=1, sharex=True, figsize=(8, 14))
            for _idx, _df in enumerate([df, df_drop]):
                # group by 'SensorID' and 'Maneuver'
                grouped_df = _df.groupby(["SensorID", "Maneuver"])
                try:
                    data_manuv = grouped_df.get_group((sensor_ID, 1)).sort_values("MJD")
                except KeyError:
                    data_manuv = grouped_df.get_group((sensor_ID, 2)).sort_values("MJD")
                data_no_manuv = grouped_df.get_group((sensor_ID, 0)).sort_values("MJD")

                axs[_ax_idx[_idx]].scatter(
                    data_manuv["Datetime"],
                    data_manuv["LON Arcsec"],
                    label="manuv=True",
                    s=m_size,
                )
                axs[_ax_idx[_idx]].scatter(
                    data_no_manuv["Datetime"],
                    data_no_manuv["LON Arcsec"],
                    label="manuv=False",
                    s=m_size,
                )
                axs[_ax_idx[_idx]].xaxis.set_major_locator(hours)
                axs[_ax_idx[_idx]].xaxis.set_major_formatter(timeFmt)
                # highlight dropouts
                if _idx == 1:
                    for drop in sensor["weather"]["drops"]:
                        axs[_ax_idx[_idx]].axvspan(
                            drop[0], drop[1], color="yellow", alpha=0.3
                        )
                axs[_ax_idx[_idx]].grid()
                axs[_ax_idx[_idx]].set_ylabel("LON Arcsec")
                axs[_ax_idx[_idx]].set_title(
                    f"Sensor_ID = {sensor_ID}; {_plt_types[_idx]}"
                )
                # -
                axs[_ax_idx[_idx] + 1].scatter(
                    data_manuv["Datetime"],
                    data_manuv["LAT Arcsec"],
                    label="manuv=True",
                    s=m_size,
                )
                axs[_ax_idx[_idx] + 1].scatter(
                    data_no_manuv["Datetime"],
                    data_no_manuv["LAT Arcsec"],
                    label="manuv=False",
                    s=m_size,
                )
                axs[_ax_idx[_idx] + 1].xaxis.set_major_locator(hours)
                axs[_ax_idx[_idx] + 1].xaxis.set_major_formatter(timeFmt)
                # highlight dropouts
                if _idx == 1:
                    for drop in sensor["weather"]["drops"]:
                        axs[_ax_idx[_idx] + 1].axvspan(
                            drop[0], drop[1], color="yellow", alpha=0.3
                        )
                        axs[_ax_idx[_idx] + 1].set_xlabel("Time (%H:)")
                axs[_ax_idx[_idx] + 1].grid()
                axs[_ax_idx[_idx] + 1].set_ylabel("LAT Arcsec")
                axs[_ax_idx[_idx] + 1].set_title(
                    f"Sensor_ID = {sensor_ID}; {_plt_types[_idx]}"
                )

            # create legend for all plots and put it outside of plot area
            handles, labels = axs[0].get_legend_handles_labels()
            fig.legend(
                handles,
                labels,
                loc="upper left",
                bbox_to_anchor=(1, 0.9),
                borderaxespad=0.0,
            )
            plt.tight_layout()

            # save plot?
            if save_without_displaying:
                outpath = Path(save_path)
                outpath.mkdir(parents=True, exist_ok=True)
                plt.savefig(
                    outpath / f"plot__sensor_id_{sensor_ID}.png", bbox_inches="tight"
                )
                plt.close()
        except KeyError:
            pass


def plot_seq_id(
    seq_id: int,
    df: pd.DataFrame,
    df_drop: pd.DataFrame,
    save_without_displaying: bool = False,
    save_path: str = "./plots/seq_id/",
) -> None:
    """This function plots the manuver=[{1,2}, 0] pair for a given seq_id.
    The plot also displays the measurements that are dropped due to a synthetic weather event.

    Parameters
    ----------
    seq_id : int
        The seq_id to plot
    df : pd.DataFrame
        The pandas dataframe of a MaDDg output file, with an additional 'Datetime' column
    df_drop : pd.DataFrame
        The pandas dataframe of a MaDDg output file, with missing data, and an additonal
        'Datetime' column
    save_without_displaying : bool, optional
        Whether or not to save the plots without displaying them, by default False
    save_path : str, optional
        Path to save the plots in the hydra multirun folder, by default "./plots/seq_id/"
    """
    idx = seq_id

    # --- make missing data df
    df_merged = pd.merge(df_drop, df, how="outer", indicator=True)
    # select the rows that are in df but not in df_drop
    df_missing = df_merged[df_merged["_merge"] == "right_only"]

    try:
        # --- group by 'Sequence' and 'Maneuver'
        grouped_df = df.groupby(["Sequence", "Maneuver"])
        grouped_df_missing = df_missing.groupby(["Sequence"])
        try:
            data_manuv = grouped_df.get_group((idx, 1)).sort_values("MJD")
        except KeyError:
            data_manuv = grouped_df.get_group((idx, 2)).sort_values("MJD")
        data_no_manuv = grouped_df.get_group((idx, 0)).sort_values("MJD")
        data_missing = grouped_df_missing.get_group(idx).sort_values("MJD")

        # --- make & display multiplot
        hours = mdates.HourLocator(interval=8)
        timeFmt = mdates.DateFormatter("%H:")
        m_size = 4
        fig, axs = plt.subplots(nrows=2, ncols=1, sharex=True)
        axs[0].scatter(
            data_manuv["Datetime"],
            data_manuv["LON Arcsec"],
            label="manuv=True",
            s=m_size,
        )
        axs[0].scatter(
            data_no_manuv["Datetime"],
            data_no_manuv["LON Arcsec"],
            label="manuv=False",
            s=m_size,
        )
        axs[0].scatter(
            data_missing["Datetime"],
            data_missing["LON Arcsec"],
            label="dropped",
            marker="x",
            color="red",
        )
        axs[0].xaxis.set_major_locator(hours)
        axs[0].xaxis.set_major_formatter(timeFmt)
        axs[0].grid()
        axs[0].set_title(f"LON Arcsec; seq_ID = {idx}")
        # -
        axs[1].scatter(
            data_manuv["Datetime"],
            data_manuv["LAT Arcsec"],
            label="manuv=True",
            s=m_size,
        )
        axs[1].scatter(
            data_no_manuv["Datetime"],
            data_no_manuv["LAT Arcsec"],
            label="manuv=False",
            s=m_size,
        )
        axs[1].scatter(
            data_missing["Datetime"],
            data_missing["LAT Arcsec"],
            label="dropped",
            marker="x",
            color="red",
        )
        axs[1].xaxis.set_major_locator(hours)
        axs[1].xaxis.set_major_formatter(timeFmt)
        axs[1].set_xlabel("Time (%H:)")
        axs[1].grid()
        axs[1].set_title(f"LAT Arcsec; seq_ID = {idx}")

        # create legent for all plots and put it outside of plot area
        handles, labels = axs[0].get_legend_handles_labels()
        fig.legend(
            handles,
            labels,
            loc="upper left",
            bbox_to_anchor=(1, 0.9),
            borderaxespad=0.0,
        )
        plt.tight_layout()

        # save plot?
        if save_without_displaying:
            outpath = Path(save_path)
            outpath.mkdir(parents=True, exist_ok=True)
            plt.savefig(outpath / f"plot__seq_id_{seq_id}.png", bbox_inches="tight")
            plt.close()
    except KeyError:
        pass


def dropout(
    path: str,
    sensor_list: List[Dict[str, Any]],
    sensor_idx: int | None = None,
    cloud_prob: float = 0.5,
    cloud_duration_mean: float = 10800.0,
    cloud_duration_std: float = 7200.0,
    num_runs: int = 1,
    save_copy_of_original: bool = True,
    save_plots: bool = False,
) -> None:
    """Applies weather related data dropout to a MaDDG data output file (.csv).

    Parameters
    ----------
    path : str
        Path to the MaDDG data output file (.csv) to apply weather related data dropout
    sensor_list : List[Dict[str, Any]]
        List of sensors
    sensor_idx : int | None, optional
        Sensor index, by default None
    cloud_prob : float, optional
        Probability of a cloud blocking a measurement: 0 - 1, by default 0.5
    cloud_duration_mean : float, optional
        Average duration of each cloud event (seconds), by default 10800.0
    cloud_duration_std : float, optional
        Standard deviation of each cloud event (seconds), by default 7200.0
    num_runs : int, optional
        Number of dropout simulations to run, by default 1
    save_copy_of_original : bool, optional
        Whether or not to include a copy of the original data in the dropout simulation
        results directory, by default True
    save_plots : bool, optional
        Whether or not to generate and save dropout plots, by default False
    """
    for num_run in range(num_runs):
        try:
            # update sensor_list params
            sensor_list = update_sensor_weather_stats(
                sensor_list=sensor_list,
                idx=sensor_idx,
                cloud_prob=cloud_prob,
                cloud_duration_mean=cloud_duration_mean,
                cloud_duration_std=cloud_duration_std,
            )

            # read data from csv
            df = pd.read_csv(path)
            ts_mjd = Time(df["MJD"], format="mjd")
            ts_datetime = ts_mjd.to_value("datetime64")
            df["Datetime"] = ts_datetime
            df.sort_values("MJD", inplace=True)
            sensor_IDs = df["SensorID"].unique()

            # group by 'SensorID'
            grouped_df = df.groupby(["SensorID"])

            seq_IDs = []
            dfs = []
            # for idx in tqdm(range(sensor_IDs.size)):
            for sensor_ID in sensor_IDs:
                # get sensor dict that corresponds to sensor_ID
                sensor = get_sensor(sensor_list, sensor_ID)
                assert sensor is not None

                # add blockout range key to sensor
                sensor["weather"]["drops"] = []

                # get subset of data associated with sensor_ID
                data = grouped_df.get_group(sensor_ID).sort_values("MJD").copy()

                # get seq_IDs in data subset
                seq_IDs = df["Sequence"].unique()

                # --- group data by day by looking for a large gap between measurements
                # set threshold for maximum gap between the elapsed times
                gap_threshold = 10000  # seconds

                # compute the gaps between the elapsed times
                gaps = (
                    (data["Datetime"] - data["Datetime"].shift(1))
                    .dt.total_seconds()
                    .values
                )

                # create labels for each contiguous group of times separated by gaps smaller than gap_threshold
                labels = np.cumsum(gaps > gap_threshold)

                # add labels to data as 'Cluster' column
                data["Cluster"] = labels

                # get unique cluster IDs
                clusters = data["Cluster"].unique()

                # group by 'Cluster'
                grouped_data = data.groupby(["Cluster"])

                # generate random dropout & duration for each "Cluster"
                rands = [np.random.rand() for _ in range(clusters.size)]
                durations = [
                    gen_random_duration(
                        loc=sensor["weather"]["cloud_duration_mean"],
                        scale=sensor["weather"]["cloud_duration_std"],
                    )
                    for _ in range(clusters.size)
                ]

                # --- weather dropout
                dfs_filtered = []
                for rand, duration, cluster in zip(rands, durations, clusters):
                    data_cluster = grouped_data.get_group(cluster).copy()
                    if (rand <= sensor["weather"]["cloud_prob"]) and (
                        cloud_duration_mean != 0
                    ):
                        # drop = True, i.e. cloud coverage event occurs
                        # log useful numbers
                        cluster_duration = (
                            data_cluster["Datetime"].max()
                            - data_cluster["Datetime"].min()
                        )
                        cloud_duration = np.timedelta64(int(duration), "s")
                        if cloud_duration > cluster_duration:
                            # mark all measurements as cloudy
                            datetime_start = data_cluster["Datetime"].min()
                            datetime_stop = data_cluster["Datetime"].max()
                        else:
                            # mark subset of measurements as cloudy for specified duration
                            datetime_start = data_cluster.sample(n=1)[
                                "Datetime"
                            ].values[0]
                            datetime_stop = datetime_start + cloud_duration
                            # make sure the end time does not exceed the latest datetime in the cluster
                            if datetime_stop > data_cluster["Datetime"].max():
                                datetime_stop = data_cluster["Datetime"].max()
                                datetime_start = datetime_stop - cloud_duration

                        mask = (data_cluster["Datetime"] >= datetime_start) & (
                            data_cluster["Datetime"] <= datetime_stop
                        )
                        data_cluster["Clouds"] = mask
                        sensor["weather"]["drops"].append(
                            [datetime_start, datetime_stop]
                        )
                    else:
                        # drop = False, i.e. no clouds tonight
                        data_cluster["Clouds"] = False

                    dfs_filtered.append(data_cluster)
                dfs.append(pd.concat(dfs_filtered, axis=0))

            # After Loop Post-process ------------------------------------------|
            dfc = pd.concat(dfs, axis=0).sort_values("MJD")
            df_drop = df.drop(dfc[dfc["Clouds"] == True].index)

            # --- save plots
            if save_plots:
                # seq ID plots:
                for seq_ID in sorted(seq_IDs):
                    plot_seq_id(
                        seq_ID,
                        df,
                        df_drop,
                        save_without_displaying=True,
                        save_path=f"./plots/num_run_{num_run}/seq_id/",
                    )  # type: ignore
                # sensor ID plots:
                for sensor_ID in sensor_IDs:
                    plot_sensor_ID(
                        sensor_ID,
                        df,
                        df_drop,
                        sensor_list,
                        save_without_displaying=True,
                        save_path=f"./plots/num_run_{num_run}/sensor_ID/",
                    )

            # save dataset csv
            df_drop.drop(columns=["Datetime"]).sort_index().to_csv(
                f"output_dropped__{num_run}.csv", index=False
            )
            # sensor_list with dropout information pickle
            pd.to_pickle(sensor_list, f"sensor_list__{num_run}.pkl")
            # copy of original, unaltered dataset
            if save_copy_of_original and (num_run == 0):
                df.drop(columns=["Datetime"]).sort_index().to_csv(
                    "output.csv", index=False
                )
        except (TypeError, KeyError):
            pass


def launcher(
    config: type[DataClass],
    task_function: Callable[[Any], Any],
    overrides: list[str],
    submitit: str = "",
    return_output_dir: bool = False,
) -> None | Path:
    """Launches a multirun hydra job via hydra-zen

    Parameters
    ----------
    config : type[DataClass]
        Hydra-zen config with user-defined field names and, optionally,
        associated default values and/or type annotations.
    task_function : Callable[[Any], Any]
        The function that Hydra will execute. Its input will be config,
        which has been modified via the specified overrides
    overrides : list[str]
        Sets/overrides values in hydra-zen ``config``
    submitit : str, optional
        If specified, the path to a config JSON file defining how to launch jobs across multiple GPUs using submitit, by default None (serial launch only)
    return_output_dir : bool, optional
        If True, the data output path will be returned (otherwise None is returned), by default False

    Returns
    -------
    None | Path
        Returns None by default, unless input argument ``return_output_dir`` is True, then the data output path will be returned

    Raises
    ------
    MadlibException
        The specified submitit configuration file (JSON) must be a list of strings
    """
    # add submitit Config overrides if submitit bool argument is True
    if submitit != "":
        with open(submitit, "r") as f:
            submitit_overrides = json.load(f)

        # Make sure the JSON contents are a list of strings
        str_list_check = all(isinstance(n, str) for n in submitit_overrides)
        if (type(submitit_overrides) != list) or (not str_list_check):
            raise MadlibException(
                f"The specified submitit configuration file {submitit} is not formatted properly. "
                "The JSON must be a list of strings."
            )

        overrides += submitit_overrides

    # launch jobs (in parallel if using submitit)
    (jobs,) = launch(
        config=config,
        task_function=task_function,
        to_dictconfig=True,
        overrides=overrides,
        version_base="1.3",
        multirun=True,
    )

    if return_output_dir:
        # get multirun root directory
        return Path(jobs[0].working_dir).parent


# %%

if __name__ == "__main__":
    # parse CLI arguments
    parser = parseArgs()
    args = parser.parse_args()

    # customize Hydra's configuration
    store(HydraConf(job=JobConf(chdir=True)))
    store.add_to_hydra_store(overwrite_ok=True)

    # Parse the sensor YAML file
    sensor_data = SensorCollection.paramsFromYAML(args.sensor_yaml)
    sensor_list = [
        sensor_data[key] for key in sensor_data.keys()
    ]  # TODO: Legacy sensor_list format. Refactor to use sensor_data dict

    # zen wrapper that will auto-extract, resolve, and instantiate fields from
    # an input config based on dropout()'s signature
    task_function = zen(dropout)

    DropoutConfig = make_config(
        path=args.path,
        sensor_list=sensor_list,
        sensor_idx=None,
        cloud_prob=0.5,
        cloud_duration_mean=10800.0,
        cloud_duration_std=3600.0,
        num_runs=args.num_runs,
        save_copy_of_original=args.save_copy_of_original,
        save_plots=args.save_plots,
    )

    overrides = [
        f"cloud_prob={args.cloud_prob}",
        f"cloud_duration_mean={args.cloud_duration_mean}",
        f"cloud_duration_std={args.cloud_duration_std}",
    ]

    launcher(
        config=DropoutConfig,
        task_function=task_function,
        overrides=overrides,
        submitit=args.submitit,
    )



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import List, Self

import numpy as np
from numpy.typing import NDArray

from ._utils import MadlibException


@dataclass(kw_only=True)
class Observation:
    """
    Class for holding observables. All angles are in degrees.

    Parameters
    ----------
    mjd : float
        Timestamp of the observation, described as a MJD in UTC

    ra : float | None
        Topocentric right ascension angle, by default None

    dec : float | None
        Topocentric declination angle, by default None

    az : float | None
        Azimuth angle, by default None

    el : float | None
        Elevation angle, by default None

    range_ : float | None
        Distance between sensor and target, by default None

    range_rate : float | None
        Time rate of change of the distance between the sensor and
        target, by default None

    lat : float | None
        Geodetic latitude, by default None

    lon : float | None
        Geodetic longitude, by default None

    sun_el : float | None
        Elevation angle of sun, by default None

    sun_separation : float | None
        Separation angle between target and sun, by default None

    sensor_id : str | None
        Unique Sensor ID, by default None
    """

    # time (MJD, UTC)
    mjd: float

    # standard observables for optics
    ra: float | None = None
    dec: float | None = None

    # standard observables for radar
    az: float | None = None
    el: float | None = None
    range_: float | None = None
    range_rate: float | None = None

    # non-standard observables
    lat: float | None = None
    lon: float | None = None

    # Position of sun
    sun_el: float | None = None
    sun_separation: float | None = None

    # Sensor ID for bookkeeping
    sensor_id: str | None = None

    _keys = [
        "ra",
        "dec",
        "az",
        "el",
        "range_",
        "range_rate",
        "lat",
        "lon",
        "sun_el",
        "sun_separation",
    ]

    def __sub__(self, other: Self) -> "ObservationResidual":
        """Subtracts another instance of Observation class.

        Parameters
        ----------
        other : Self
            The other instance of Observation class to subract

        Returns
        -------
        Self
            The subtracted result

        Raises
        ------
        MadlibException
            Can only subtract two Observation objects
        MadlibException
            Observations must be at the same time for computing a residual
        """

        if not isinstance(other, Observation):
            raise MadlibException("Can only subtract two Observation objects")

        d1 = self.__dict__
        d2 = other.__dict__
        diff = {
            key: d1[key] - d2[key]
            for key in self._keys
            if d1[key] is not None and d2[key] is not None
        }

        # special handling of angles that wrap at 0/360
        for key in ("ra", "az", "lon"):
            if d1[key] is not None and d2[key] is not None:
                temp = np.unwrap(np.array([d1[key], d2[key]]), period=360)
                diff[key] = temp[0] - temp[1]

        # add the timestamp to the dict
        if abs(d1["mjd"] - d2["mjd"]) > 1e-9:
            raise MadlibException(
                "Observations must be at the same time for computing a residual"
            )

        diff["mjd"] = d1["mjd"]

        # Remove the solar elevation
        _ = diff.pop("sun_el", None)

        return ObservationResidual(**diff)

    def asarray(self) -> NDArray[np.float64]:
        """Convert this observation to a flat 1-D array"""
        return np.array(
            [val if val is not None else np.NaN for val in self.__dict__.values()]
        )


@dataclass(kw_only=True)
class ObservationResidual:
    """
    Class for holding the difference between two observables.

    Parameters
    ----------
    mjd : float
        Timestamp of the observation, described as a MJD in UTC

    ra : float | None
        Topocentric right ascension angle difference, by default None

    dec : float | None
        Topocentric declination angle difference, by default None

    az : float | None
        Azimuth angle difference, by default None

    el : float | None
        Elevation angle difference, by default None

    range_ : float | None
        Distance between sensor and target difference, by default None

    range_rate : float | None
        Time rate of change of the distance between the sensor and target
        difference, by default None

    lat : float | None
        Geodetic latitude difference, by default None

    lon : float | None
        Geodetic longitude difference, by default None
    """

    # time (MJD, UTC)
    mjd: float

    # standard observables for optics
    ra: float | None = None
    dec: float | None = None

    # standard observables for radar
    az: float | None = None
    el: float | None = None
    range_: float | None = None
    range_rate: float | None = None

    # non-standard observables
    lat: float | None = None
    lon: float | None = None

    def asarray(self) -> NDArray[np.float64]:
        """Convert this observation to a flat 1-D array"""
        return np.array(
            [val if val is not None else np.NaN for val in self.__dict__.values()]
        )


@dataclass(kw_only=True)
class ObservationCollection:
    """Class for holding observed and true positions of satellites.

    Parameters
    ----------
    pos_observed: np.ndarray[Observation, np.dtype[np.float64]]
        Realistic observations of a satellite given sensor noise parameters

    pos_truth: np.ndarray[Observation, np.dtype[np.float64]]
        True observations of a satellite ignoring all noise sources

    pos_expected: np.ndarray[Observation, np.dtype[np.float64]]
        Observations expected if no noise or maneuvers occur

    Raises
    ------
    MadlibException
        Can only add two ObservationCollection objects
    """

    pos_observed: np.ndarray[Observation, np.dtype[np.float64]]
    pos_truth: np.ndarray[Observation, np.dtype[np.float64]]
    pos_expected: np.ndarray[Observation, np.dtype[np.float64]]

    def __add__(self, other: "ObservationCollection"):
        self.pos_observed = np.concatenate((self.pos_observed, other.pos_observed))
        self.pos_truth = np.concatenate((self.pos_truth, other.pos_truth))
        self.pos_expected = np.concatenate((self.pos_expected, other.pos_expected))

    def count_valid_observations(self):
        return len(self.pos_observed)


def combineObsCollections(
    collectionList: List[ObservationCollection],
) -> ObservationCollection:
    """Given observations of a satellite from multiple sensors, combine them into a single object.

    Parameters
    ----------
    collectionList : List[ObservationCollection]
        List of observations from multiple sensors of a single satellite.

    Returns
    -------
    ObservationCollection
        Combined collection of observations from all sensors.
    """

    pos_observed = np.concatenate([col.pos_observed for col in collectionList])
    pos_truth = np.concatenate([col.pos_truth for col in collectionList])
    pos_expected = np.concatenate([col.pos_expected for col in collectionList])

    combinedCollection = ObservationCollection(
        pos_observed=pos_observed,
        pos_truth=pos_truth,
        pos_expected=pos_expected,
    )

    return combinedCollection



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd
from astropy.time import Time

import madlib


def calculate_residuals(
    sensors: madlib.SensorCollection,
    satellite: madlib.Satellite,
    sim_duration_days: float,
    t_start_mjd: float = Time.now().mjd,
) -> pd.DataFrame | None:
    """Calculates the residuals.

    Parameters
    ----------
    sensors : madlib.SensorCollection
        Collection of Sensors
    satellite : madlib.Satellite
        The Satellite object observed in the simulation
    sim_duration_days : float
        Duration of the simulation (days)
    t_start_mjd : float, optional
        Time (MJD) at the start of the simulation, by default astropy.time.Time.now().mjd

    Returns
    -------
    pd.DataFrame | None
        pandas DataFrame of results (or None, if nobs == 0)

    Raises
    ------
    MadlibException
        Raised if nobs != nexpected...
        number of actual and expected observations were not
        the same, so residuals cannot be calculated. This
        is likely a random occurrence.
    """
    ### SIMULATION TIMING

    t_end_mjd = t_start_mjd + sim_duration_days
    sensors.generate_obs_timing(t_start_mjd, t_end_mjd)

    ### GENERATE OBSERVATIONS
    obs = sensors.observe(satellite)

    actual_obs = obs.pos_observed
    predicted_obs = obs.pos_expected

    nobs = obs.count_valid_observations()

    ### CALCULATE AND RETURN RESIDUALS
    output = None
    if nobs > 0:
        obs_res = np.array([actual_obs[n] - predicted_obs[n] for n in range(nobs)])
        ra_res = np.array([obs_res[n].ra for n in range(nobs)]) * 3600
        dec_res = np.array([obs_res[n].dec for n in range(nobs)]) * 3600
        t = np.array([x.mjd for x in actual_obs])
        sensor_ids = np.array([x.sensor_id for x in predicted_obs])

        output = {
            "MJD": t,
            "RA Arcsec": ra_res,
            "DEC Arcsec": dec_res,
            "SensorID": sensor_ids,
        }
        output = pd.DataFrame(output)

    return output



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Test file: test_hz_launcher.py
Description: This file contains unit tests for the `_sim_launcher` module
"""

from maddg._sim_launcher import launcher
from scripts.hz_launcher import simulator_task
from pathlib import Path
import pandas as pd
import numpy as np
import shutil
from madlib._utils import MadlibException
import yaml
from unittest import mock

sensor_yaml_11 = "configs/sample_sensor_network.yaml"
sensor_yaml_blind = "tests/inputs/blind_sensor.yaml"


def clear_outputs():
    """Delete contents of tests/outputs"""
    outputs_dir = Path("tests/outputs")
    if outputs_dir.is_dir():
        for f in outputs_dir.glob("*"):
            f.unlink()


def count_experiments(multirun_dir: str):
    """Given a directory containing hydra-zen multirun output, count the number
    of experiments that the directory currently contains."""
    multirun_path = Path(multirun_dir)
    experiment_count = 0
    experiment_dirs = []
    if multirun_path.is_dir():
        experiment_dirs = list(multirun_path.glob("*/*"))
        experiment_count = len(experiment_dirs)

    return experiment_count, experiment_dirs


def clear_experiment(experiment_dir: Path):
    """Given an experiment directory, delete it and its contents. If its
    parent is then empty, delete the parent. Same for grandparent."""
    parent = experiment_dir.parent
    grandparent = parent.parent
    shutil.rmtree(experiment_dir)

    siblings = list(parent.glob("*"))
    if len(siblings) == 0:
        shutil.rmtree(parent)

        aunts = list(grandparent.glob("*"))
        if len(aunts) == 0:
            shutil.rmtree(grandparent)


class TestLauncher:
    """Test the behavior of the launcher function"""

    def test_impulsive(self):
        """Test launching experiments with impulsive maneuvers"""

        # This test should generate files and directories under ./multirun,
        # as well as summary outputs under ./tests/outputs. We'll start by
        # clearing out the ./tests/outputs directory and measuring how many
        # experiments already exist under ./multirun
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 3
        start_mjd = 60196.5

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_11,
            outdir=outdir,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        # The number of multirun experiments should have increased by 1
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count + 1

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert output_csv.exists()
        assert errors_txt.exists()

        results = pd.read_csv(output_csv)

        assert results["Sequence"].isin([0, 1]).all()
        assert len(results) > 0
        assert sum(results["Maneuver"] == 0) > 0
        assert sum(results["Maneuver"] == 1) > 0

        no_mnvr = results.loc[results["Maneuver"] == 0]
        mnvr = results.loc[results["Maneuver"] == 1]

        assert all(no_mnvr["Maneuver_Start_MJD"].isna())
        assert all(no_mnvr["Maneuver_End_MJD"].isna())
        assert all(no_mnvr["Maneuver_DV_Radial_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_InTrack_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_CrossTrack_KmS"].isna())

        assert all(mnvr["Maneuver_Start_MJD"] > start_mjd)
        assert all(mnvr["Maneuver_Start_MJD"] < start_mjd + sim_duration_days)
        assert all(mnvr["Maneuver_End_MJD"].isna())
        np.testing.assert_allclose(mnvr["Maneuver_DV_Radial_KmS"], 0, atol=1e-6)
        assert all(np.abs(mnvr["Maneuver_DV_InTrack_KmS"]) > 1e-6)
        assert all(np.abs(mnvr["Maneuver_DV_CrossTrack_KmS"]) > 1e-6)

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # Cleanup
        clear_outputs()
        experiment_dir = (set(new_experiment_dirs) - set(experiment_dirs)).pop()
        clear_experiment(experiment_dir)

    def test_impulsive_with_modified_sensor_dra_ddec(self):
        """Test launching experiments with impulsive maneuvers"""

        # This test should generate files and directories under ./multirun,
        # as well as summary outputs under ./tests/outputs. We'll start by
        # clearing out the ./tests/outputs directory and measuring how many
        # experiments already exist under ./multirun
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 3
        start_mjd = 60196.5

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_11,
            outdir=outdir,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            random_seed=0,
            sensor_ddec=12.0,
            sensor_dra=7.7,
        )

        # The number of multirun experiments should have increased by 1
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count + 1

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")
        multirun_yaml_path = Path("tests/outputs/multirun.yaml")

        assert output_csv.exists()
        assert errors_txt.exists()

        results = pd.read_csv(output_csv)

        assert results["Sequence"].isin([0, 1]).all()
        assert len(results) > 0
        assert sum(results["Maneuver"] == 0) > 0
        assert sum(results["Maneuver"] == 1) > 0

        no_mnvr = results.loc[results["Maneuver"] == 0]
        mnvr = results.loc[results["Maneuver"] == 1]

        assert all(no_mnvr["Maneuver_Start_MJD"].isna())
        assert all(no_mnvr["Maneuver_End_MJD"].isna())
        assert all(no_mnvr["Maneuver_DV_Radial_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_InTrack_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_CrossTrack_KmS"].isna())

        assert all(no_mnvr["Maneuver_Start_MJD"].isna())
        assert all(no_mnvr["Maneuver_End_MJD"].isna())
        assert all(no_mnvr["Maneuver_DV_Radial_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_InTrack_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_CrossTrack_KmS"].isna())

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # check `dra` and `ddec` were changed
        with open(multirun_yaml_path, "r") as file:
            multirun_yaml = yaml.safe_load(file)

        multirun_yaml_sensor_dras = [
            multirun_yaml["sensor_params"][sensor]["dra"]
            for sensor in multirun_yaml["sensor_params"].keys()
        ]
        multirun_yaml_sensor_ddecs = [
            multirun_yaml["sensor_params"][sensor]["ddec"]
            for sensor in multirun_yaml["sensor_params"].keys()
        ]
        np.testing.assert_allclose(multirun_yaml_sensor_dras, 7.7, atol=1e-6)
        np.testing.assert_allclose(multirun_yaml_sensor_ddecs, 12.0, atol=1e-6)

        # Cleanup
        clear_outputs()
        experiment_dir = (set(new_experiment_dirs) - set(experiment_dirs)).pop()
        clear_experiment(experiment_dir)

    def test_continuous(self):
        """Test launching experiments with continuous maneuvers"""

        # This test should generate files and directories under ./multirun,
        # as well as summary outputs under ./tests/outputs. We'll start by
        # clearing out the ./tests/outputs directory and measuring how many
        # experiments already exist under ./multirun
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "continuous"
        sim_duration_days = 0.5
        start_mjd = 60196.5
        cont_thrust_duration_days = 2.0
        cont_thrust_mag = 1e-5
        cont_thrust_model = 0

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_11,
            outdir=outdir,
            cont_thrust_duration_days=cont_thrust_duration_days,
            cont_thrust_mag=cont_thrust_mag,
            cont_thrust_model=cont_thrust_model,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        # The number of multirun experiments should have increased by 1
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count + 1

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert output_csv.exists()
        assert errors_txt.exists()

        results = pd.read_csv(output_csv)

        assert results["Sequence"].isin([0, 2]).all()
        assert len(results) > 0
        assert sum(results["Maneuver"] == 0) > 0
        assert sum(results["Maneuver"] == 2) > 0

        no_mnvr = results.loc[results["Maneuver"] == 0]
        mnvr = results.loc[results["Maneuver"] == 2]

        assert all(no_mnvr["Maneuver_Start_MJD"].isna())
        assert all(no_mnvr["Maneuver_End_MJD"].isna())
        assert all(no_mnvr["Maneuver_DV_Radial_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_InTrack_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_CrossTrack_KmS"].isna())

        np.testing.assert_allclose(mnvr["Maneuver_Start_MJD"], start_mjd, atol=1e-6)
        np.testing.assert_allclose(
            mnvr["Maneuver_End_MJD"], start_mjd + cont_thrust_duration_days, atol=1e-6
        )
        np.testing.assert_allclose(mnvr["Maneuver_DV_Radial_KmS"], 0, atol=1e-6)
        np.testing.assert_allclose(mnvr["Maneuver_DV_InTrack_KmS"], 1e-5, atol=1e-6)
        np.testing.assert_allclose(mnvr["Maneuver_DV_CrossTrack_KmS"], 0, atol=1e-6)

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # Cleanup
        clear_outputs()
        experiment_dir = (set(new_experiment_dirs) - set(experiment_dirs)).pop()
        clear_experiment(experiment_dir)

    def test_no_obs(self):
        """Test that a simulation with no observations creates an empty csv"""
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs_blank"
        mtype = "impulse"
        sim_duration_days = 0.5
        start_mjd = 60196.5

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_blind,
            outdir=outdir,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            random_seed=0,
        )

        # The number of multirun experiments should have increased by 1
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count + 1

        output_csv = Path(outdir) / "complete.csv"
        errors_txt = Path(outdir) / "errors.txt"

        assert output_csv.exists()
        assert errors_txt.exists()

        with open(output_csv, "r") as f:
            output = f.read()

        assert output.strip() == ""

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # Cleanup
        clear_outputs()
        shutil.rmtree(outdir)
        experiment_dir = (set(new_experiment_dirs) - set(experiment_dirs)).pop()
        clear_experiment(experiment_dir)

    def test_auto_cleanup(self):
        """Test that the rm_multirun_root option will clean up multirun directories."""

        # This test should generate files and directories under ./multirun,
        # as well as summary outputs under ./tests/outputs. We'll start by
        # clearing out the ./tests/outputs directory and measuring how many
        # experiments already exist under ./multirun
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 0.5
        start_mjd = 60196.5

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_11,
            outdir=outdir,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            rm_multirun_root=True,
            random_seed=0,
        )

        # The number of multirun experiments should not have changed
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert output_csv.exists()
        assert errors_txt.exists()

        results = pd.read_csv(output_csv)

        assert results["Sequence"].isin([0, 1]).all()
        assert len(results) > 0
        assert sum(results["Maneuver"] == 0) > 0
        assert sum(results["Maneuver"] == 1) > 0

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # Cleanup
        clear_outputs()

    def test_prediction_error(self):
        """Test launching experiments with orbit-misestimation"""

        # This test should generate files and directories under ./multirun,
        # as well as summary outputs under ./tests/outputs. We'll start by
        # clearing out the ./tests/outputs directory and measuring how many
        # experiments already exist under ./multirun
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 3
        start_mjd = 60196.5

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_11,
            outdir=outdir,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            random_seed=0,
            pred_err=0.1,
        )

        # The number of multirun experiments should have increased by 1
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count + 1

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert output_csv.exists()
        assert errors_txt.exists()

        results = pd.read_csv(output_csv)

        assert results["Sequence"].isin([0, 1]).all()
        assert len(results) > 0
        assert sum(results["Maneuver"] == 0) > 0
        assert sum(results["Maneuver"] == 1) > 0

        no_mnvr = results.loc[results["Maneuver"] == 0]
        mnvr = results.loc[results["Maneuver"] == 1]

        assert all(no_mnvr["Maneuver_Start_MJD"].isna())
        assert all(no_mnvr["Maneuver_End_MJD"].isna())
        assert all(no_mnvr["Maneuver_DV_Radial_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_InTrack_KmS"].isna())
        assert all(no_mnvr["Maneuver_DV_CrossTrack_KmS"].isna())

        assert all(mnvr["Maneuver_Start_MJD"] > start_mjd)
        assert all(mnvr["Maneuver_Start_MJD"] < start_mjd + sim_duration_days)
        assert all(mnvr["Maneuver_End_MJD"].isna())
        np.testing.assert_allclose(mnvr["Maneuver_DV_Radial_KmS"], 0, atol=1e-6)
        assert all(np.abs(mnvr["Maneuver_DV_InTrack_KmS"]) > 1e-6)
        assert all(np.abs(mnvr["Maneuver_DV_CrossTrack_KmS"]) > 1e-6)

        # Make sure that even the non-maneuver case has large residuals
        assert np.abs(no_mnvr["RA Arcsec"].iloc[0]) > 1000.0
        assert np.abs(no_mnvr["DEC Arcsec"].iloc[0]) > 1000.0

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # Cleanup
        clear_outputs()
        experiment_dir = (set(new_experiment_dirs) - set(experiment_dirs)).pop()
        clear_experiment(experiment_dir)

    def test_auto_cleanup_with_error(self):
        """Test that the rm_multirun_root option will clean up multirun directories
        even if an error.txt file is produced."""

        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 3
        start_mjd = 60196.5

        invalid_sensor_yaml = "tests/inputs/invalid_sensor_4.yaml"

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=invalid_sensor_yaml,
            outdir=outdir,
            dv_ric_mean_kms=(1000, 1000, 1000),
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            rm_multirun_root=True,
            random_seed=0,
        )

        # The number of multirun experiments should not have changed
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert output_csv.exists()
        assert errors_txt.exists()

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors != ""

        # Cleanup
        clear_outputs()

    def test_cleanup_without_hydra_dir(self):
        """Make sure nothing breaks if the .hydra folder does not exist
        when it's time to delete it."""

        original_exists = Path.exists

        # We'll need to mock Path.exists() to make this test work
        def mock_exists(self):
            if self.stem == ".hydra":
                return False
            else:
                return original_exists(self)

        with mock.patch.object(Path, "exists", mock_exists):
            clear_outputs()
            experiment_count, experiment_dirs = count_experiments("multirun")

            num_sim_pairs = 1
            outdir = "tests/outputs"
            mtype = "impulse"
            sim_duration_days = 3
            start_mjd = 60196.5

            invalid_sensor_yaml = "tests/inputs/invalid_sensor_4.yaml"

            launcher(
                simulator_method=simulator_task,
                mtype=mtype,
                num_sim_pairs=num_sim_pairs,
                sensor_yaml=invalid_sensor_yaml,
                outdir=outdir,
                dv_ric_mean_kms=(1000, 1000, 1000),
                start_mjd=start_mjd,
                sim_duration_days=sim_duration_days,
                rm_multirun_root=True,
                random_seed=0,
            )

            # The number of multirun experiments should have increased
            new_experiment_count, new_experiment_dirs = count_experiments("multirun")
            assert new_experiment_count == experiment_count + 1

            output_csv = Path("tests/outputs/complete.csv")
            errors_txt = Path("tests/outputs/errors.txt")

            assert output_csv.exists()
            assert errors_txt.exists()

            with open(errors_txt, "r") as f:
                errors = f.read()

            assert errors != ""

            # Cleanup
            clear_outputs()

    def test_auto_cleanup_custom_root(self):
        """Test that the rm_multirun_root option will clean up multirun directories."""
        clear_outputs()
        multirun_root = "tests/outputs/multirun"
        experiment_count, experiment_dirs = count_experiments(multirun_root)

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 0.5
        start_mjd = 60196.5

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_11,
            outdir=outdir,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            multirun_root=multirun_root,
            rm_multirun_root=True,
            random_seed=0,
        )

        # The number of multirun experiments should not have changed
        new_experiment_count, new_experiment_dirs = count_experiments(multirun_root)
        assert new_experiment_count == experiment_count

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert output_csv.exists()
        assert errors_txt.exists()

        results = pd.read_csv(output_csv)

        assert results["Sequence"].isin([0, 1]).all()
        assert len(results) > 0
        assert sum(results["Maneuver"] == 0) > 0
        assert sum(results["Maneuver"] == 1) > 0

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # Cleanup
        clear_outputs()

    def test_auto_cleanup_partial(self):
        """Check that directories are not deleted if they contain a non-hydra file."""
        clear_outputs()
        multirun_root = "tests/outputs/multirun_partial"
        experiment_count, experiment_dirs = count_experiments(multirun_root)

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 0.5
        start_mjd = 60196.5

        original_exists = Path.exists

        # We'll need to mock Path.exists() to make this test work
        def mock_exists(self):
            if self.stem == ".hydra":
                # Create new files to prevent complete erasure
                files = list(self.glob("../*"))
                for f in files:
                    newfile = Path(f"{f}.blank")
                    with open(newfile, "w") as f:
                        f.write("")

            return original_exists(self)

        with mock.patch.object(Path, "exists", mock_exists):

            launcher(
                simulator_method=simulator_task,
                mtype=mtype,
                num_sim_pairs=num_sim_pairs,
                sensor_yaml=sensor_yaml_11,
                outdir=outdir,
                start_mjd=start_mjd,
                sim_duration_days=sim_duration_days,
                multirun_root=multirun_root,
                rm_multirun_root=True,
                random_seed=0,
            )

            # The number of multirun experiments should have increased
            new_experiment_count, new_experiment_dirs = count_experiments(multirun_root)
            assert new_experiment_count == experiment_count + 1

            output_csv = Path("tests/outputs/complete.csv")
            errors_txt = Path("tests/outputs/errors.txt")

            assert output_csv.exists()
            assert errors_txt.exists()

            with open(errors_txt, "r") as f:
                errors = f.read()

            assert errors == ""

            # Cleanup
            experiment_dir = (set(new_experiment_dirs) - set(experiment_dirs)).pop()
            clear_experiment(experiment_dir)
            clear_outputs()

    def test_yaml_copy_exception(self):
        """Check that code still runs as expected if multirun.yaml is not found at the end."""
        clear_outputs()
        multirun_root = "tests/outputs_noyaml/multirun_noyaml"
        experiment_count, experiment_dirs = count_experiments(multirun_root)

        num_sim_pairs = 1
        outdir = "tests/outputs_noyaml"
        mtype = "impulse"
        sim_duration_days = 0.5
        start_mjd = 60196.5

        if Path(outdir).exists() and Path(outdir).is_dir():
            shutil.rmtree(Path(outdir))

        original_glob = Path.glob

        # We'll need to mock Path.exists() to make this test work
        def mock_glob(self, args):
            if args == "multirun.yaml":
                (self / "multirun.yaml").unlink(missing_ok=True)
            return original_glob(self, args)

        with mock.patch.object(Path, "glob", mock_glob):
            launcher(
                simulator_method=simulator_task,
                mtype=mtype,
                num_sim_pairs=num_sim_pairs,
                sensor_yaml=sensor_yaml_11,
                outdir=outdir,
                start_mjd=start_mjd,
                sim_duration_days=sim_duration_days,
                multirun_root=multirun_root,
                rm_multirun_root=True,
                random_seed=0,
            )

            # The number of multirun experiments should not have increased
            new_experiment_count, new_experiment_dirs = count_experiments(multirun_root)
            assert new_experiment_count == experiment_count

            # The copied multirun.yaml should not exist
            assert not (Path(outdir) / "multirun.yaml").exists()

            output_csv = Path(outdir) / "complete.csv"
            errors_txt = Path(outdir) / "errors.txt"

            assert output_csv.exists()
            assert errors_txt.exists()

            with open(errors_txt, "r") as f:
                errors = f.read()

            assert errors == ""

            # Cleanup
            clear_outputs()
            shutil.rmtree(outdir)


class TestSubmitit:
    """These tests will check the behavior of the hydra submitit launcher"""

    def test_submitit(self):
        """Test that the submitit overrides work (NOTE: This will not actually run submitit)"""
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 0.5
        start_mjd = 60196.5

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_11,
            outdir=outdir,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            rm_multirun_root=True,
            random_seed=0,
            submitit="tests/inputs/submitit_test.json",
        )

        # The number of multirun experiments should not have changed
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert output_csv.exists()
        assert errors_txt.exists()

        results = pd.read_csv(output_csv)

        assert results["Sequence"].isin([0, 1]).all()
        assert len(results) > 0
        assert sum(results["Maneuver"] == 0) > 0
        assert sum(results["Maneuver"] == 1) > 0

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # Cleanup
        clear_outputs()

    def test_invalid(self):
        """Test that a submitit JSON with invalid format will be rejected"""
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 0.5
        start_mjd = 60196.5

        failed = False
        try:
            launcher(
                simulator_method=simulator_task,
                mtype=mtype,
                num_sim_pairs=num_sim_pairs,
                sensor_yaml=sensor_yaml_11,
                outdir=outdir,
                start_mjd=start_mjd,
                sim_duration_days=sim_duration_days,
                rm_multirun_root=True,
                random_seed=0,
                submitit="tests/inputs/submitit_invalid.json",
            )
        except MadlibException:
            failed = True

        assert failed

        # The number of multirun experiments should not have changed
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert not output_csv.exists()
        assert not errors_txt.exists()

    def test_auto_cleanup(self):
        """Test that the rm_multirun_root option will clean up multirun directories."""

        # This test should generate files and directories under ./multirun,
        # as well as summary outputs under ./tests/outputs. We'll start by
        # clearing out the ./tests/outputs directory and measuring how many
        # experiments already exist under ./multirun
        clear_outputs()
        experiment_count, experiment_dirs = count_experiments("multirun")

        num_sim_pairs = 1
        outdir = "tests/outputs"
        mtype = "impulse"
        sim_duration_days = 0.5
        start_mjd = 60196.5

        launcher(
            simulator_method=simulator_task,
            mtype=mtype,
            num_sim_pairs=num_sim_pairs,
            sensor_yaml=sensor_yaml_11,
            outdir=outdir,
            start_mjd=start_mjd,
            sim_duration_days=sim_duration_days,
            rm_multirun_root=True,
            random_seed=0,
            submitit="tests/inputs/submitit_test.json",
        )

        # The number of multirun experiments should not have changed
        new_experiment_count, new_experiment_dirs = count_experiments("multirun")
        assert new_experiment_count == experiment_count

        output_csv = Path("tests/outputs/complete.csv")
        errors_txt = Path("tests/outputs/errors.txt")

        assert output_csv.exists()
        assert errors_txt.exists()

        results = pd.read_csv(output_csv)

        assert results["Sequence"].isin([0, 1]).all()
        assert len(results) > 0
        assert sum(results["Maneuver"] == 0) > 0
        assert sum(results["Maneuver"] == 1) > 0

        with open(errors_txt, "r") as f:
            errors = f.read()

        assert errors == ""

        # Cleanup
        clear_outputs()
